@@ -10,10 +10,46 @@
 **********************************************************************/
 
 #include "ruby/ruby.h"
+#include "internal.h"
 #include "vm_core.h"
 #include "gc.h"
 #include "eval_intern.h"
 
+#if ((defined(_WIN32) && _WIN32_WINNT >= 0x0400) || (defined(HAVE_GETCONTEXT) && defined(HAVE_SETCONTEXT))) && !defined(__NetBSD__) && !defined(sun) && !defined(FIBER_USE_NATIVE)
+#define FIBER_USE_NATIVE 1
+
+/* FIBER_USE_NATIVE enables Fiber performance improvement using system
+ * dependent method such as make/setcontext on POSIX system or
+ * CreateFiber() API on Windows.
+ * This hack make Fiber context switch faster (x2 or more).
+ * However, it decrease maximum number of Fiber.  For example, on the
+ * 32bit POSIX OS, ten or twenty thousands Fiber can be created.
+ *
+ * Details is reported in the paper "A Fast Fiber Implementation for Ruby 1.9"
+ * in Proc. of 51th Programming Symposium, pp.21--28 (2010) (in Japanese).
+ */
+
+/* On our experience, NetBSD doesn't support using setcontext() and pthread
+ * simultaneously.  This is because pthread_self(), TLS and other information
+ * are represented by stack pointer (higher bits of stack pointer).
+ * TODO: check such constraint on configure.
+ */
+#elif !defined(FIBER_USE_NATIVE)
+#define FIBER_USE_NATIVE 0
+#endif
+
+#if FIBER_USE_NATIVE
+#ifndef _WIN32
+#include <unistd.h>
+#include <sys/mman.h>
+#include <ucontext.h>
+#endif
+#define RB_PAGE_SIZE (pagesize)
+#define RB_PAGE_MASK (~(RB_PAGE_SIZE - 1))
+static long pagesize;
+#define FIBER_MACHINE_STACK_ALLOCATION_SIZE  (0x10000)
+#endif
+
 #define CAPTURE_JUST_VALID_VM_STACK 1
 
 enum context_type {
@@ -50,12 +86,30 @@ enum fiber_status {
     TERMINATED
 };
 
+#if FIBER_USE_NATIVE && !defined(_WIN32)
+#define MAX_MAHINE_STACK_CACHE  10
+static int machine_stack_cache_index = 0;
+typedef struct machine_stack_cache_struct {
+    void *ptr;
+    size_t size;
+} machine_stack_cache_t;
+static machine_stack_cache_t machine_stack_cache[MAX_MAHINE_STACK_CACHE];
+static machine_stack_cache_t terminated_machine_stack;
+#endif
+
 typedef struct rb_fiber_struct {
     rb_context_t cont;
     VALUE prev;
     enum fiber_status status;
     struct rb_fiber_struct *prev_fiber;
     struct rb_fiber_struct *next_fiber;
+#if FIBER_USE_NATIVE
+#ifdef _WIN32
+    void *fib_handle;
+#else
+    ucontext_t context;
+#endif
+#endif
 } rb_fiber_t;
 
 static const rb_data_type_t cont_data_type, fiber_data_type;
@@ -64,18 +118,17 @@ static VALUE rb_cFiber;
 static VALUE rb_eFiberError;
 
 #define GetContPtr(obj, ptr)  \
-    TypedData_Get_Struct(obj, rb_context_t, &cont_data_type, ptr)
+    TypedData_Get_Struct((obj), rb_context_t, &cont_data_type, (ptr))
 
 #define GetFiberPtr(obj, ptr)  do {\
-    TypedData_Get_Struct(obj, rb_fiber_t, &fiber_data_type, ptr); \
-    if (!ptr) rb_raise(rb_eFiberError, "uninitialized fiber"); \
+    TypedData_Get_Struct((obj), rb_fiber_t, &fiber_data_type, (ptr)); \
+    if (!(ptr)) rb_raise(rb_eFiberError, "uninitialized fiber"); \
 } while(0)
 
 NOINLINE(static VALUE cont_capture(volatile int *stat));
 
-void rb_thread_mark(rb_thread_t *th);
 #define THREAD_MUST_BE_RUNNING(th) do { \
-	if (!th->tag) rb_raise(rb_eThreadError, "not running thread");	\
+	if (!(th)->tag) rb_raise(rb_eThreadError, "not running thread");	\
     } while (0)
 
 static void
@@ -86,6 +139,7 @@ cont_mark(void *ptr)
 	rb_context_t *cont = ptr;
 	rb_gc_mark(cont->value);
 	rb_thread_mark(&cont->saved_thread);
+	rb_gc_mark(cont->saved_thread.self);
 
 	if (cont->vm_stack) {
 #ifdef CAPTURE_JUST_VALID_VM_STACK
@@ -98,8 +152,21 @@ cont_mark(void *ptr)
 	}
 
 	if (cont->machine_stack) {
-	    rb_gc_mark_locations(cont->machine_stack,
-				 cont->machine_stack + cont->machine_stack_size);
+	    if (cont->type == CONTINUATION_CONTEXT) {
+		/* cont */
+		rb_gc_mark_locations(cont->machine_stack,
+				     cont->machine_stack + cont->machine_stack_size);
+            }
+            else {
+		/* fiber */
+		rb_thread_t *th;
+                rb_fiber_t *fib = (rb_fiber_t*)cont;
+		GetThreadPtr(cont->saved_thread.self, th);
+		if ((th->fiber != cont->self) && fib->status == RUNNING) {
+		    rb_gc_mark_locations(cont->machine_stack,
+					 cont->machine_stack + cont->machine_stack_size);
+		}
+	    }
 	}
 #ifdef __ia64
 	if (cont->machine_register_stack) {
@@ -118,7 +185,41 @@ cont_free(void *ptr)
     if (ptr) {
 	rb_context_t *cont = ptr;
 	RUBY_FREE_UNLESS_NULL(cont->saved_thread.stack); fflush(stdout);
+#if FIBER_USE_NATIVE
+	if (cont->type == CONTINUATION_CONTEXT) {
+	    /* cont */
+	    RUBY_FREE_UNLESS_NULL(cont->machine_stack);
+	}
+	else {
+	    /* fiber */
+#ifdef _WIN32
+	    if (GET_THREAD()->fiber != cont->self && cont->type != ROOT_FIBER_CONTEXT) {
+		/* don't delete root fiber handle */
+		rb_fiber_t *fib = (rb_fiber_t*)cont;
+		if (fib->fib_handle) {
+		    DeleteFiber(fib->fib_handle);
+		}
+	    }
+#else /* not WIN32 */
+	    if (GET_THREAD()->fiber != cont->self) {
+                rb_fiber_t *fib = (rb_fiber_t*)cont;
+                if (fib->context.uc_stack.ss_sp) {
+                    if (cont->type == ROOT_FIBER_CONTEXT) {
+			rb_bug("Illegal root fiber parameter");
+                    }
+		    munmap((void*)fib->context.uc_stack.ss_sp, fib->context.uc_stack.ss_size);
+		}
+	    }
+            else {
+		/* It may reached here when finalize */
+		/* TODO examine whether it is a bug */
+                /* rb_bug("cont_free: release self"); */
+            }
+#endif
+	}
+#else /* not FIBER_USE_NATIVE */
 	RUBY_FREE_UNLESS_NULL(cont->machine_stack);
+#endif
 #ifdef __ia64
 	RUBY_FREE_UNLESS_NULL(cont->machine_register_stack);
 #endif
@@ -197,7 +298,6 @@ fiber_free(void *ptr)
     RUBY_FREE_ENTER("fiber");
     if (ptr) {
 	rb_fiber_t *fib = ptr;
-
 	if (fib->cont.type != ROOT_FIBER_CONTEXT &&
 	    fib->cont.saved_thread.local_storage) {
 	    st_free_table(fib->cont.saved_thread.local_storage);
@@ -228,7 +328,6 @@ static void
 cont_save_machine_stack(rb_thread_t *th, rb_context_t *cont)
 {
     size_t size;
-    rb_thread_t *sth = &cont->saved_thread;
 
     SET_MACHINE_STACK_END(&th->machine_stack_end);
 #ifdef __ia64
@@ -267,25 +366,34 @@ cont_save_machine_stack(rb_thread_t *th, rb_context_t *cont)
 
     MEMCPY(cont->machine_register_stack, cont->machine_register_stack_src, VALUE, size);
 #endif
-
-    sth->machine_stack_start = sth->machine_stack_end = 0;
-#ifdef __ia64
-    sth->machine_register_stack_start = sth->machine_register_stack_end = 0;
-#endif
 }
 
 static const rb_data_type_t cont_data_type = {
     "continuation",
-    cont_mark, cont_free, cont_memsize,
+    {cont_mark, cont_free, cont_memsize,},
 };
 
+static void
+cont_save_thread(rb_context_t *cont, rb_thread_t *th)
+{
+    /* save thread context */
+    cont->saved_thread = *th;
+    /* saved_thread->machine_stack_(start|end) should be NULL */
+    /* because it may happen GC afterward */
+    cont->saved_thread.machine_stack_start = 0;
+    cont->saved_thread.machine_stack_end = 0;
+#ifdef __ia64
+    cont->saved_thread.machine_register_stack_start = 0;
+    cont->saved_thread.machine_register_stack_end = 0;
+#endif
+}
+
 static void
 cont_init(rb_context_t *cont, rb_thread_t *th)
 {
     /* save thread context */
-    cont->saved_thread = *th;
+    cont_save_thread(cont, th);
     cont->saved_thread.local_storage = 0;
-    cont->saved_thread.machine_stack_start = cont->saved_thread.machine_stack_end = 0;
 }
 
 static rb_context_t *
@@ -302,8 +410,6 @@ cont_new(VALUE klass)
     return cont;
 }
 
-void rb_vm_stack_to_heap(rb_thread_t *th);
-
 static VALUE
 cont_capture(volatile int *stat)
 {
@@ -332,7 +438,7 @@ cont_capture(volatile int *stat)
     cont_save_machine_stack(th, cont);
 
     if (ruby_setjmp(cont->jmpbuf)) {
-	VALUE value;
+	volatile VALUE value;
 
 	value = cont->value;
 	if (cont->argc == -1) rb_exc_raise(value);
@@ -346,10 +452,8 @@ cont_capture(volatile int *stat)
     }
 }
 
-NOINLINE(NORETURN(static void cont_restore_1(rb_context_t *)));
-
 static void
-cont_restore_1(rb_context_t *cont)
+cont_restore_thread(rb_context_t *cont)
 {
     rb_thread_t *th = GET_THREAD(), *sth = &cont->saved_thread;
 
@@ -392,6 +496,179 @@ cont_restore_1(rb_context_t *cont)
     th->protect_tag = sth->protect_tag;
     th->errinfo = sth->errinfo;
     th->first_proc = sth->first_proc;
+}
+
+#if FIBER_USE_NATIVE
+#ifdef _WIN32
+static void
+fiber_set_stack_location(void)
+{
+    rb_thread_t *th = GET_THREAD();
+    VALUE *ptr;
+
+    SET_MACHINE_STACK_END(&ptr);
+    th->machine_stack_start = (void*)(((VALUE)ptr & RB_PAGE_MASK) + STACK_UPPER((void *)&ptr, 0, RB_PAGE_SIZE));
+}
+
+static VOID CALLBACK
+fiber_entry(void *arg)
+{
+    fiber_set_stack_location();
+    rb_fiber_start();
+}
+#else /* _WIN32 */
+
+/*
+ * FreeBSD require a first (i.e. addr) argument of mmap(2) is not NULL
+ * if MAP_STACK is passed.
+ * http://www.FreeBSD.org/cgi/query-pr.cgi?pr=158755
+ */
+#if defined(MAP_STACK) && !defined(__FreeBSD__) && !defined(__FreeBSD_kernel__)
+#define FIBER_STACK_FLAGS (MAP_PRIVATE | MAP_ANON | MAP_STACK)
+#else
+#define FIBER_STACK_FLAGS (MAP_PRIVATE | MAP_ANON)
+#endif
+
+static char*
+fiber_machine_stack_alloc(size_t size)
+{
+    char *ptr;
+
+    if (machine_stack_cache_index > 0) {
+	if (machine_stack_cache[machine_stack_cache_index - 1].size == (size / sizeof(VALUE))) {
+	    ptr = machine_stack_cache[machine_stack_cache_index - 1].ptr;
+	    machine_stack_cache_index--;
+	    machine_stack_cache[machine_stack_cache_index].ptr = NULL;
+	    machine_stack_cache[machine_stack_cache_index].size = 0;
+	}
+	else{
+            /* TODO handle multiple machine stack size */
+	    rb_bug("machine_stack_cache size is not canonicalized");
+	}
+    }
+    else {
+	void *page;
+	STACK_GROW_DIR_DETECTION;
+
+	ptr = mmap(NULL, size, PROT_READ | PROT_WRITE, FIBER_STACK_FLAGS, -1, 0);
+	if (ptr == MAP_FAILED) {
+	    rb_raise(rb_eFiberError, "can't alloc machine stack to fiber");
+	}
+
+	/* guard page setup */
+	page = ptr + STACK_DIR_UPPER(size - RB_PAGE_SIZE, 0);
+	if (mprotect(page, RB_PAGE_SIZE, PROT_NONE) < 0) {
+	    rb_raise(rb_eFiberError, "mprotect failed");
+	}
+    }
+
+    return ptr;
+}
+#endif
+
+static void
+fiber_initialize_machine_stack_context(rb_fiber_t *fib, size_t size)
+{
+    rb_thread_t *sth = &fib->cont.saved_thread;
+
+#ifdef _WIN32
+    fib->fib_handle = CreateFiberEx(size - 1, size, 0, fiber_entry, NULL);
+    if (!fib->fib_handle) {
+	/* try to release unnecessary fibers & retry to create */
+	rb_gc();
+	fib->fib_handle = CreateFiberEx(size - 1, size, 0, fiber_entry, NULL);
+	if (!fib->fib_handle) {
+	    rb_raise(rb_eFiberError, "can't create fiber");
+	}
+    }
+    sth->machine_stack_maxsize = size;
+#else /* not WIN32 */
+    ucontext_t *context = &fib->context;
+    char *ptr;
+    STACK_GROW_DIR_DETECTION;
+
+    getcontext(context);
+    ptr = fiber_machine_stack_alloc(size);
+    context->uc_link = NULL;
+    context->uc_stack.ss_sp = ptr;
+    context->uc_stack.ss_size = size;
+    makecontext(context, rb_fiber_start, 0);
+    sth->machine_stack_start = (VALUE*)(ptr + STACK_DIR_UPPER(0, size));
+    sth->machine_stack_maxsize = size - RB_PAGE_SIZE;
+#endif
+#ifdef __ia64
+    sth->machine_register_stack_maxsize = sth->machine_stack_maxsize;
+#endif
+}
+
+NOINLINE(static void fiber_setcontext(rb_fiber_t *newfib, rb_fiber_t *oldfib));
+
+static void
+fiber_setcontext(rb_fiber_t *newfib, rb_fiber_t *oldfib)
+{
+    rb_thread_t *th = GET_THREAD(), *sth = &newfib->cont.saved_thread;
+
+    if (newfib->status != RUNNING) {
+	fiber_initialize_machine_stack_context(newfib, FIBER_MACHINE_STACK_ALLOCATION_SIZE);
+    }
+
+    /* restore thread context */
+    cont_restore_thread(&newfib->cont);
+    th->machine_stack_maxsize = sth->machine_stack_maxsize;
+    if (sth->machine_stack_end && (newfib != oldfib)) {
+	rb_bug("fiber_setcontext: sth->machine_stack_end has non zero value");
+    }
+
+    /* save  oldfib's machine stack */
+    if (oldfib->status != TERMINATED) {
+	STACK_GROW_DIR_DETECTION;
+	SET_MACHINE_STACK_END(&th->machine_stack_end);
+	if (STACK_DIR_UPPER(0, 1)) {
+	    oldfib->cont.machine_stack_size = th->machine_stack_start - th->machine_stack_end;
+	    oldfib->cont.machine_stack = th->machine_stack_end;
+	}
+	else {
+	    oldfib->cont.machine_stack_size = th->machine_stack_end - th->machine_stack_start;
+	    oldfib->cont.machine_stack = th->machine_stack_start;
+	}
+    }
+    /* exchange machine_stack_start between oldfib and newfib */
+    oldfib->cont.saved_thread.machine_stack_start = th->machine_stack_start;
+    th->machine_stack_start = sth->machine_stack_start;
+    /* oldfib->machine_stack_end should be NULL */
+    oldfib->cont.saved_thread.machine_stack_end = 0;
+#ifndef _WIN32
+    if (!newfib->context.uc_stack.ss_sp && th->root_fiber != newfib->cont.self) {
+	rb_bug("non_root_fiber->context.uc_stac.ss_sp should not be NULL");
+    }
+#endif
+
+    /* swap machine context */
+#ifdef _WIN32
+    SwitchToFiber(newfib->fib_handle);
+#elif defined(__FreeBSD__) /* FreeBSD 9 doesn't work with swapcontext */
+    if (!ruby_setjmp(oldfib->cont.jmpbuf)) {
+	if (newfib->status != RUNNING) {
+	    if (setcontext(&newfib->context) < 0) {
+		rb_bug("context switch between fiber failed");
+	    }
+	}
+	else {
+	    ruby_longjmp(newfib->cont.jmpbuf, 1);
+	}
+    }
+#else
+    swapcontext(&oldfib->context, &newfib->context);
+#endif
+}
+#endif
+
+NOINLINE(NORETURN(static void cont_restore_1(rb_context_t *)));
+
+static void
+cont_restore_1(rb_context_t *cont)
+{
+    cont_restore_thread(cont);
 
     /* restore machine stack */
 #ifdef _M_AMD64
@@ -406,7 +683,7 @@ cont_restore_1(rb_context_t *cont)
     if (cont->machine_stack_src) {
 	FLUSH_REGISTER_WINDOWS;
 	MEMCPY(cont->machine_stack_src, cont->machine_stack,
-	       VALUE, cont->machine_stack_size);
+		VALUE, cont->machine_stack_size);
     }
 
 #ifdef __ia64
@@ -505,24 +782,25 @@ cont_restore_0(rb_context_t *cont, VALUE *addr_in_prev_frame)
     cont_restore_1(cont);
 }
 #ifdef __ia64
-#define cont_restore_0(cont, vp) register_stack_extend(cont, vp, (VALUE*)rb_ia64_bsp());
+#define cont_restore_0(cont, vp) register_stack_extend((cont), (vp), (VALUE*)rb_ia64_bsp());
 #endif
 
 /*
  *  Document-class: Continuation
  *
- *  Continuation objects are generated by
- *  <code>Kernel#callcc</code>. They hold a return address and execution
- *  context, allowing a nonlocal return to the end of the
- *  <code>callcc</code> block from anywhere within a program.
- *  Continuations are somewhat analogous to a structured version of C's
- *  <code>setjmp/longjmp</code> (although they contain more state, so
- *  you might consider them closer to threads).
+ *  Continuation objects are generated by <code>Kernel#callcc</code>,
+ *  after having <code>require</code>d <i>continuation</i>. They hold
+ *  a return address and execution context, allowing a nonlocal return
+ *  to the end of the <code>callcc</code> block from anywhere within a
+ *  program.  Continuations are somewhat analogous to a structured
+ *  version of C's <code>setjmp/longjmp</code> (although they contain
+ *  more state, so you might consider them closer to threads).
  *
  *  For instance:
  *
+ *     require "continuation"
  *     arr = [ "Freddie", "Herbie", "Ron", "Max", "Ringo" ]
- *     callcc{|$cc|}
+ *     callcc{|cc| $cc = cc}
  *     puts(message = arr.shift)
  *     $cc.call unless message =~ /Max/
  *
@@ -536,6 +814,7 @@ cont_restore_0(rb_context_t *cont, VALUE *addr_in_prev_frame)
  *  This (somewhat contrived) example allows the inner loop to abandon
  *  processing early:
  *
+ *     require "continuation"
  *     callcc {|cont|
  *       for i in 0..4
  *         print "\n#{i}: "
@@ -545,7 +824,7 @@ cont_restore_0(rb_context_t *cont, VALUE *addr_in_prev_frame)
  *         end
  *       end
  *     }
- *     print "\n"
+ *     puts
  *
  *  <em>produces:</em>
  *
@@ -559,14 +838,16 @@ cont_restore_0(rb_context_t *cont, VALUE *addr_in_prev_frame)
  *  call-seq:
  *     callcc {|cont| block }   ->  obj
  *
- *  Generates a <code>Continuation</code> object, which it passes to the
- *  associated block. Performing a <em>cont</em><code>.call</code> will
- *  cause the <code>callcc</code> to return (as will falling through the
- *  end of the block). The value returned by the <code>callcc</code> is
- *  the value of the block, or the value passed to
- *  <em>cont</em><code>.call</code>. See class <code>Continuation</code>
- *  for more details. Also see <code>Kernel::throw</code> for
- *  an alternative mechanism for unwinding a call stack.
+ *  Generates a <code>Continuation</code> object, which it passes to
+ *  the associated block. You need to <code>require
+ *  'continuation'</code> before using this method. Performing a
+ *  <em>cont</em><code>.call</code> will cause the <code>callcc</code>
+ *  to return (as will falling through the end of the block). The
+ *  value returned by the <code>callcc</code> is the value of the
+ *  block, or the value passed to <em>cont</em><code>.call</code>. See
+ *  class <code>Continuation</code> for more details. Also see
+ *  <code>Kernel::throw</code> for an alternative mechanism for
+ *  unwinding a call stack.
  */
 
 static VALUE
@@ -711,7 +992,7 @@ rb_cont_call(int argc, VALUE *argv, VALUE contval)
 
 static const rb_data_type_t fiber_data_type = {
     "fiber",
-    fiber_mark, fiber_free, fiber_memsize,
+    {fiber_mark, fiber_free, fiber_memsize,},
 };
 
 static VALUE
@@ -726,6 +1007,10 @@ fiber_t_alloc(VALUE fibval)
     rb_fiber_t *fib;
     rb_thread_t *th = GET_THREAD();
 
+    if (DATA_PTR(fibval) != 0) {
+	rb_raise(rb_eRuntimeError, "cannot initialize twice");
+    }
+
     THREAD_MUST_BE_RUNNING(th);
     fib = ALLOC(rb_fiber_t);
     memset(fib, 0, sizeof(rb_fiber_t));
@@ -747,7 +1032,6 @@ fiber_init(VALUE fibval, VALUE proc)
     rb_context_t *cont = &fib->cont;
     rb_thread_t *th = &cont->saved_thread;
 
-
     /* initialize cont */
     cont->vm_stack = 0;
 
@@ -778,7 +1062,9 @@ fiber_init(VALUE fibval, VALUE proc)
 
     th->first_proc = proc;
 
+#if !FIBER_USE_NATIVE
     MEMCPY(&cont->jmpbuf, &th->root_jmpbuf, rb_jmpbuf_t, 1);
+#endif
 
     return fibval;
 }
@@ -827,6 +1113,14 @@ rb_fiber_terminate(rb_fiber_t *fib)
 {
     VALUE value = fib->cont.value;
     fib->status = TERMINATED;
+#if FIBER_USE_NATIVE && !defined(_WIN32)
+    /* Ruby must not switch to other thread until storing terminated_machine_stack */
+    terminated_machine_stack.ptr = fib->context.uc_stack.ss_sp;
+    terminated_machine_stack.size = fib->context.uc_stack.ss_size / sizeof(VALUE);
+    fib->context.uc_stack.ss_sp = NULL;
+    fib->cont.machine_stack = NULL;
+    fib->cont.machine_stack_size = 0;
+#endif
     rb_fiber_transfer(return_fiber(), 1, &value);
 }
 
@@ -878,10 +1172,15 @@ static rb_fiber_t *
 root_fiber_alloc(rb_thread_t *th)
 {
     rb_fiber_t *fib;
-
     /* no need to allocate vm stack */
     fib = fiber_t_alloc(fiber_alloc(rb_cFiber));
     fib->cont.type = ROOT_FIBER_CONTEXT;
+#if FIBER_USE_NATIVE
+#ifdef _WIN32
+    fib->fib_handle = ConvertThreadToFiber(0);
+#endif
+#endif
+    fib->status = RUNNING;
     fib->prev_fiber = fib->next_fiber = fib;
 
     return fib;
@@ -907,7 +1206,7 @@ fiber_store(rb_fiber_t *next_fib)
 
     if (th->fiber) {
 	GetFiberPtr(th->fiber, fib);
-	fib->cont.saved_thread = *th;
+	cont_save_thread(&fib->cont, th);
     }
     else {
 	/* create current fiber */
@@ -915,17 +1214,43 @@ fiber_store(rb_fiber_t *next_fib)
 	th->root_fiber = th->fiber = fib->cont.self;
     }
 
+#if !FIBER_USE_NATIVE
     cont_save_machine_stack(th, &fib->cont);
 
     if (ruby_setjmp(fib->cont.jmpbuf)) {
+#else /* FIBER_USE_NATIVE */
+    {
+	fiber_setcontext(next_fib, fib);
+#ifndef _WIN32
+	if (terminated_machine_stack.ptr) {
+	    if (machine_stack_cache_index < MAX_MAHINE_STACK_CACHE) {
+		machine_stack_cache[machine_stack_cache_index].ptr = terminated_machine_stack.ptr;
+		machine_stack_cache[machine_stack_cache_index].size = terminated_machine_stack.size;
+		machine_stack_cache_index++;
+	    }
+	    else {
+		if (terminated_machine_stack.ptr != fib->cont.machine_stack) {
+		    munmap((void*)terminated_machine_stack.ptr, terminated_machine_stack.size * sizeof(VALUE));
+		}
+		else {
+		    rb_bug("terminated fiber resumed");
+		}
+	    }
+	    terminated_machine_stack.ptr = NULL;
+	    terminated_machine_stack.size = 0;
+	}
+#endif
+#endif
 	/* restored */
 	GetFiberPtr(th->fiber, fib);
 	if (fib->cont.argc == -1) rb_exc_raise(fib->cont.value);
 	return fib->cont.value;
     }
+#if !FIBER_USE_NATIVE
     else {
 	return Qundef;
     }
+#endif
 }
 
 static inline VALUE
@@ -960,7 +1285,17 @@ fiber_switch(VALUE fibval, int argc, VALUE *argv, int is_resume)
 	cont = &fib->cont;
 	cont->argc = -1;
 	cont->value = value;
+#if FIBER_USE_NATIVE
+	{
+	    VALUE oldfibval;
+	    rb_fiber_t *oldfib;
+	    oldfibval = rb_fiber_current();
+	    GetFiberPtr(oldfibval, oldfib);
+	    fiber_setcontext(fib, oldfib);
+	}
+#else
 	cont_restore_0(cont, &value);
+#endif
     }
 
     if (is_resume) {
@@ -970,11 +1305,13 @@ fiber_switch(VALUE fibval, int argc, VALUE *argv, int is_resume)
     cont->argc = argc;
     cont->value = make_passing_arg(argc, argv);
 
-    if ((value = fiber_store(fib)) == Qundef) {
+    value = fiber_store(fib);
+#if !FIBER_USE_NATIVE
+    if (value == Qundef) {
 	cont_restore_0(cont, &value);
 	rb_bug("rb_fiber_resume: unreachable");
     }
-
+#endif
     RUBY_VM_CHECK_INTS();
 
     return value;
@@ -992,7 +1329,7 @@ rb_fiber_resume(VALUE fibval, int argc, VALUE *argv)
     rb_fiber_t *fib;
     GetFiberPtr(fibval, fib);
 
-    if (fib->prev != Qnil) {
+    if (fib->prev != Qnil || fib->cont.type == ROOT_FIBER_CONTEXT) {
 	rb_raise(rb_eFiberError, "double resume");
     }
 
@@ -1009,9 +1346,10 @@ rb_fiber_yield(int argc, VALUE *argv)
  *  call-seq:
  *     fiber.alive? -> true or false
  *
- *  Returns true if the fiber can still be resumed (or transferred to).
- *  After finishing execution of the fiber block this method will always
- *  return false.
+ *  Returns true if the fiber can still be resumed (or transferred
+ *  to). After finishing execution of the fiber block this method will
+ *  always return false. You need to <code>require 'fiber'</code>
+ *  before using this method.
  */
 VALUE
 rb_fiber_alive_p(VALUE fibval)
@@ -1048,7 +1386,9 @@ rb_fiber_m_resume(int argc, VALUE *argv, VALUE fib)
  *
  *  Transfer control to another fiber, resuming it from where it last
  *  stopped or starting it if it was not resumed before. The calling
- *  fiber will be suspended much like in a call to <code>Fiber.yield</code>.
+ *  fiber will be suspended much like in a call to
+ *  <code>Fiber.yield</code>. You need to <code>require 'fiber'</code>
+ *  before using this method.
  *
  *  The fiber which receives the transfer call is treats it much like
  *  a resume call. Arguments passed to transfer are treated like those
@@ -1112,6 +1452,19 @@ rb_fiber_s_current(VALUE klass)
 void
 Init_Cont(void)
 {
+#if FIBER_USE_NATIVE
+    rb_thread_t *th = GET_THREAD();
+
+#ifdef _WIN32
+    SYSTEM_INFO info;
+    GetSystemInfo(&info);
+    pagesize = info.dwPageSize;
+#else /* not WIN32 */
+    pagesize = sysconf(_SC_PAGESIZE);
+#endif
+    SET_MACHINE_STACK_END(&th->machine_stack_end);
+#endif
+
     rb_cFiber = rb_define_class("Fiber", rb_cObject);
     rb_define_alloc_func(rb_cFiber, fiber_alloc);
     rb_eFiberError = rb_define_class("FiberError", rb_eStandardError);
@@ -1120,6 +1473,10 @@ Init_Cont(void)
     rb_define_method(rb_cFiber, "resume", rb_fiber_m_resume, -1);
 }
 
+#if defined __GNUC__ && __GNUC__ >= 4
+#pragma GCC visibility push(default)
+#endif
+
 void
 ruby_Init_Continuation_body(void)
 {
@@ -1138,3 +1495,7 @@ ruby_Init_Fiber_as_Coroutine(void)
     rb_define_method(rb_cFiber, "alive?", rb_fiber_alive_p, 0);
     rb_define_singleton_method(rb_cFiber, "current", rb_fiber_s_current, 0);
 }
+
+#if defined __GNUC__ && __GNUC__ >= 4
+#pragma GCC visibility pop
+#endif
