@@ -2,13 +2,13 @@
 
   readline.c - GNU Readline module
 
-  $Author: yugui $
+  $Author: usa $
   created at: Wed Jan 20 13:59:32 JST 1999
 
   Copyright (C) 1997-2008  Shugo Maeda
-  Copyright (C) 2008       TAKAO Kouji
+  Copyright (C) 2008-2009  TAKAO Kouji
 
-  $Id: readline.c 22513 2009-02-22 09:47:41Z yugui $
+  $Id: readline.c 39377 2013-02-22 05:27:47Z usa $
 
   Contact:
    - TAKAO Kouji <kouji at takao7 dot net> (current maintainer)
@@ -40,13 +40,27 @@
 #include <unistd.h>
 #endif
 
+#ifdef HAVE_SYS_STAT_H
+#include <sys/stat.h>
+#endif
+
 static VALUE mReadline;
 
 #define EDIT_LINE_LIBRARY_VERSION "EditLine wrapper"
+#ifndef USE_INSERT_IGNORE_ESCAPE
+# if !defined(HAVE_EDITLINE_READLINE_H) && defined(HAVE_RL_PROMPT_START_IGNORE) && defined(HAVE_RL_PROMPT_END_IGNORE)
+#  define USE_INSERT_IGNORE_ESCAPE 1
+# else
+#  define USE_INSERT_IGNORE_ESCAPE 0
+# endif
+#endif
 
 #define COMPLETION_PROC "completion_proc"
 #define COMPLETION_CASE_FOLD "completion_case_fold"
 static ID completion_proc, completion_case_fold;
+#if USE_INSERT_IGNORE_ESCAPE
+static ID id_orig_prompt, id_last_prompt;
+#endif
 
 #ifndef HAVE_RL_FILENAME_COMPLETION_FUNCTION
 # define rl_filename_completion_function filename_completion_function
@@ -59,17 +73,19 @@ static ID completion_proc, completion_case_fold;
 #endif
 
 static int (*history_get_offset_func)(int);
+static int (*history_replace_offset_func)(int);
+#ifdef HAVE_RL_COMPLETION_APPEND_CHARACTER
+static int readline_completion_append_character;
+#endif
 
 static char **readline_attempted_completion_function(const char *text,
                                                      int start, int end);
 
 #define OutputStringValue(str) do {\
     SafeStringValue(str);\
-    str = rb_str_conv_enc(str, rb_enc_get(str), rb_locale_encoding());\
+    (str) = rb_str_conv_enc((str), rb_enc_get(str), rb_locale_encoding());\
 } while (0)\
 
-#ifdef HAVE_RL_EVENT_HOOK
-#define BUSY_WAIT 0
 
 /*
  * Document-class: Readline
@@ -83,7 +99,7 @@ static char **readline_attempted_completion_function(const char *text,
  * GNU Readline:: http://www.gnu.org/directory/readline.html
  * libedit::      http://www.thrysoee.dk/editline/
  *
- * Reads one inputted line with line edit by Readline.readline method. 
+ * Reads one inputted line with line edit by Readline.readline method.
  * At this time, the facilitatation completion and the key
  * bind like Emacs can be operated like GNU Readline.
  *
@@ -106,6 +122,58 @@ static char **readline_attempted_completion_function(const char *text,
  * Documented by TAKAO Kouji <kouji at takao7 dot net>.
  */
 
+#if defined HAVE_RL_GETC_FUNCTION
+static VALUE readline_instream;
+static ID id_getbyte;
+
+#ifndef HAVE_RL_GETC
+#define rl_getc(f) EOF
+#endif
+
+static int readline_getc(FILE *);
+static int
+readline_getc(FILE *input)
+{
+    rb_io_t *ifp = 0;
+    VALUE c;
+    if (!readline_instream) return rl_getc(input);
+    GetOpenFile(readline_instream, ifp);
+    if (rl_instream != ifp->stdio_file) return rl_getc(input);
+#if defined(_WIN32)
+    {
+        INPUT_RECORD ir;
+        int n;
+        static int prior_key = '0';
+        for (;;) {
+            if (prior_key > 0xff) {
+                prior_key = rl_getc(ifp->stdio_file);
+                return prior_key;
+            }
+            if (PeekConsoleInput((HANDLE)_get_osfhandle(ifp->fd), &ir, 1, &n)) {
+                if (n == 1) {
+                    if (ir.EventType == KEY_EVENT && ir.Event.KeyEvent.bKeyDown) {
+                        prior_key = rl_getc(ifp->stdio_file);
+                        return prior_key;
+                    } else {
+                        ReadConsoleInput((HANDLE)_get_osfhandle(ifp->fd), &ir, 1, &n);
+                    }
+                } else {
+                    HANDLE h = (HANDLE)_get_osfhandle(ifp->fd);
+                    rb_w32_wait_events(&h, 1, INFINITE);
+                }
+            } else {
+                break;
+            }
+        }
+    }
+#endif
+    c = rb_funcall(readline_instream, id_getbyte, 0, 0);
+    if (NIL_P(c)) return EOF;
+    return NUM2CHR(c);
+}
+#elif defined HAVE_RL_EVENT_HOOK
+#define BUSY_WAIT 0
+
 static int readline_event(void);
 static int
 readline_event(void)
@@ -113,16 +181,96 @@ readline_event(void)
 #if BUSY_WAIT
     rb_thread_schedule();
 #else
-    fd_set rset;
-
-    FD_ZERO(&rset);
-    FD_SET(fileno(rl_instream), &rset);
-    rb_thread_select(fileno(rl_instream) + 1, &rset, NULL, NULL, NULL);
+    rb_wait_for_single_fd(fileno(rl_instream), RB_WAITFD_IN, NULL);
     return 0;
 #endif
 }
 #endif
 
+#if USE_INSERT_IGNORE_ESCAPE
+static VALUE
+insert_ignore_escape(VALUE self, VALUE prompt)
+{
+    VALUE last_prompt, orig_prompt = rb_attr_get(self, id_orig_prompt);
+    int ignoring = 0;
+    const char *s0, *s, *e;
+    long len;
+    static const char ignore_code[2] = {RL_PROMPT_START_IGNORE, RL_PROMPT_END_IGNORE};
+
+    prompt = rb_str_new_shared(prompt);
+    last_prompt = rb_attr_get(self, id_last_prompt);
+    if (orig_prompt == prompt) return last_prompt;
+    len = RSTRING_LEN(prompt);
+    if (NIL_P(last_prompt)) {
+	last_prompt = rb_str_tmp_new(len);
+    }
+
+    s = s0 = RSTRING_PTR(prompt);
+    e = s0 + len;
+    rb_str_set_len(last_prompt, 0);
+    while (s < e && *s) {
+	switch (*s) {
+	  case RL_PROMPT_START_IGNORE:
+	    ignoring = -1;
+	    rb_str_cat(last_prompt, s0, ++s - s0);
+	    s0 = s;
+	    break;
+	  case RL_PROMPT_END_IGNORE:
+	    ignoring = 0;
+	    rb_str_cat(last_prompt, s0, ++s - s0);
+	    s0 = s;
+	    break;
+	  case '\033':
+	    if (++s < e && *s == '[') {
+		rb_str_cat(last_prompt, s0, s - s0 - 1);
+		s0 = s - 1;
+		while (++s < e && *s) {
+		    if (ISALPHA(*s)) {
+			if (!ignoring) {
+			    ignoring = 1;
+			    rb_str_cat(last_prompt, ignore_code+0, 1);
+			}
+			rb_str_cat(last_prompt, s0, ++s - s0);
+			s0 = s;
+			break;
+		    }
+		    else if (!('0' <= *s && *s <= '9' || *s == ';')) {
+			break;
+		    }
+		}
+	    }
+	    break;
+	  default:
+	    if (ignoring > 0) {
+		ignoring = 0;
+		rb_str_cat(last_prompt, ignore_code+1, 1);
+	    }
+	    s++;
+	    break;
+	}
+    }
+    if (ignoring > 0) {
+	ignoring = 0;
+	rb_str_cat(last_prompt, ignore_code+1, 1);
+    }
+    rb_str_cat(last_prompt, s0, s - s0);
+
+    rb_ivar_set(self, id_orig_prompt, prompt);
+    rb_ivar_set(self, id_last_prompt, last_prompt);
+
+    return last_prompt;
+}
+#endif
+
+static VALUE
+readline_get(VALUE prompt)
+{
+#ifdef HAVE_RL_COMPLETION_APPEND_CHARACTER
+    readline_completion_append_character = rl_completion_append_character;
+#endif
+    return (VALUE)readline((char *)prompt);
+}
+
 /*
  * call-seq:
  *   Readline.readline(prompt = "", add_hist = false) -> string or nil
@@ -149,9 +297,9 @@ readline_event(void)
  *
  * * Catches the Interrupt exception by pressed ^C after returns
  *   terminal status:
- * 
+ *
  *     require "readline"
- *     
+ *
  *     stty_save = `stty -g`.chomp
  *     begin
  *       while buf = Readline.readline
@@ -163,25 +311,25 @@ readline_event(void)
  *         end
  *       end
  *     end
- * 
+ *
  * * Catches the INT signal by pressed ^C after returns terminal
  *   status:
- * 
+ *
  *     require "readline"
- *     
+ *
  *     stty_save = `stty -g`.chomp
  *     trap("INT") { system "stty", stty_save; exit }
- *   
+ *
  *     while buf = Readline.readline
  *       p buf
  *     end
  *
  * * Ignores pressing ^C:
- * 
+ *
  *     require "readline"
- *     
+ *
  *     trap("INT", "SIG_IGN")
- *     
+ *
  *     while buf = Readline.readline
  *       p buf
  *     end
@@ -191,18 +339,18 @@ readline_event(void)
  * the same it as last one.
  *
  *   require "readline"
- *     
+ *
  *   while buf = Readline.readline("> ", true)
  *     # p Readline::HISTORY.to_a
  *     Readline::HISTORY.pop if /^\s*$/ =~ buf
- *  
+ *
  *     begin
  *       if Readline::HISTORY[Readline::HISTORY.length-2] == buf
- *         Readline::HISTORY.pop 
+ *         Readline::HISTORY.pop
  *       end
  *     rescue IndexError
  *     end
- *  
+ *
  *     # p Readline::HISTORY.to_a
  *     print "-> ", buf, "\n"
  *   end
@@ -220,16 +368,37 @@ readline_readline(int argc, VALUE *argv, VALUE self)
     rb_secure(4);
     if (rb_scan_args(argc, argv, "02", &tmp, &add_hist) > 0) {
 	OutputStringValue(tmp);
+#if USE_INSERT_IGNORE_ESCAPE
+	tmp = insert_ignore_escape(self, tmp);
+	rb_str_locktmp(tmp);
+#endif
 	prompt = RSTRING_PTR(tmp);
     }
 
-    if (!isatty(0) && errno == EBADF) rb_raise(rb_eIOError, "closed stdin");
+    if (!isatty(fileno(rl_instream)) && errno == EBADF) rb_raise(rb_eIOError, "closed stdin");
+    if (rl_outstream) {
+	struct stat stbuf;
+	int fd = fileno(rl_outstream);
+	if (fd < 0 || fstat(fd, &stbuf) != 0) {
+	    rb_raise(rb_eIOError, "closed stdout");
+	}
+    }
 
-    buff = (char*)rb_protect((VALUE(*)_((VALUE)))readline, (VALUE)prompt,
-                              &status);
+#ifdef _WIN32
+    rl_prep_terminal(1);
+#endif
+    buff = (char*)rb_protect(readline_get, (VALUE)prompt, &status);
+#if USE_INSERT_IGNORE_ESCAPE
+    if (prompt) {
+	rb_str_unlocktmp(tmp);
+    }
+#endif
     if (status) {
 #if defined HAVE_RL_CLEANUP_AFTER_SIGNAL
         /* restore terminal mode and signal handler*/
+#if defined HAVE_RL_FREE_LINE_STATE
+        rl_free_line_state();
+#endif
         rl_cleanup_after_signal();
 #elif defined HAVE_RL_DEPREP_TERM_FUNCTION
         /* restore terminal mode */
@@ -272,6 +441,9 @@ readline_s_set_input(VALUE self, VALUE input)
     Check_Type(input, T_FILE);
     GetOpenFile(input, ifp);
     rl_instream = rb_io_stdio_file(ifp);
+#ifdef HAVE_RL_GETC_FUNCTION
+    readline_instream = input;
+#endif
     return input;
 }
 
@@ -300,19 +472,77 @@ readline_s_set_output(VALUE self, VALUE output)
  * call-seq:
  *   Readline.completion_proc = proc
  *
- * Specifies a Proc object +proc+ to determine completion behavior. It
- * should take input-string, and return an array of completion
- * candidates.
+ * Specifies a Proc object +proc+ to determine completion behavior.  It
+ * should take input string and return an array of completion candidates.
  *
- * Raises ArgumentError exception if +proc+ does not respond to call method.
+ * The default completion is used if +proc+ is nil.
  *
- * Raises SecurityError exception if $SAFE is 4.
+ * The String that is passed to the Proc depends on the
+ * Readline.completer_word_break_characters property.  By default the word
+ * under the cursor is passed to the Proc.  For example, if the input is "foo
+ * bar" then only "bar" would be passed to the completion Proc.
+ *
+ * Upon successful completion the Readline.completion_append_character will be
+ * appended to the input so the user can start working on their next argument.
+ *
+ * = Examples
+ *
+ * == Completion for a Static List
+ *
+ *   require 'readline'
+ *
+ *   LIST = [
+ *     'search', 'download', 'open',
+ *     'help', 'history', 'quit',
+ *     'url', 'next', 'clear',
+ *     'prev', 'past'
+ *   ].sort
+ *
+ *   comp = proc { |s| LIST.grep(/^#{Regexp.escape(s)}/) }
+ *
+ *   Readline.completion_append_character = " "
+ *   Readline.completion_proc = comp
+ *
+ *   while line = Readline.readline('> ', true)
+ *     p line
+ *   end
+ *
+ * == Completion For Directory Contents
+ *
+ *   require 'readline'
+ *
+ *   Readline.completion_append_character = " "
+ *   Readline.completion_proc = Proc.new do |str|
+ *     Dir[str+'*'].grep(/^#{Regexp.escape(str)}/)
+ *   end
+ *
+ *   while line = Readline.readline('> ', true)
+ *     p line
+ *   end
+ *
+ * = Autocomplete strategies
+ *
+ * When working with auto-complete there are some strategies that work well.
+ * To get some ideas you can take a look at the
+ * completion.rb[http://svn.ruby-lang.org/repos/ruby/trunk/lib/irb/completion.rb]
+ * file for irb.
+ *
+ * The common strategy is to take a list of possible completions and filter it
+ * down to those completions that start with the user input.  In the above
+ * examples Enumerator.grep is used.  The input is escaped to prevent Regexp
+ * special characters from interfering with the matching.
+ *
+ * It may also be helpful to use the Abbrev library to generate completions.
+ *
+ * Raises ArgumentError if +proc+ does not respond to the call method.
+ *
+ * Raises SecurityError if $SAFE is 4.
  */
 static VALUE
 readline_s_set_completion_proc(VALUE self, VALUE proc)
 {
     rb_secure(4);
-    if (!rb_respond_to(proc, rb_intern("call")))
+    if (!NIL_P(proc) && !rb_respond_to(proc, rb_intern("call")))
 	rb_raise(rb_eArgError, "argument must respond to `call'");
     return rb_ivar_set(mReadline, completion_proc, proc);
 }
@@ -357,7 +587,7 @@ readline_s_set_completion_case_fold(VALUE self, VALUE val)
  * Readline.completion_case_fold= method.
  *
  *   require "readline"
- *   
+ *
  *   Readline.completion_case_fold = "This is a String."
  *   p Readline.completion_case_fold # => "This is a String."
  *
@@ -370,31 +600,89 @@ readline_s_get_completion_case_fold(VALUE self)
     return rb_attr_get(mReadline, completion_case_fold);
 }
 
+#ifdef HAVE_RL_LINE_BUFFER
+/*
+ * call-seq:
+ *   Readline.line_buffer -> string
+ *
+ * Returns the full line that is being edited. This is useful from
+ * within the complete_proc for determining the context of the
+ * completion request.
+ *
+ * The length of +Readline.line_buffer+ and GNU Readline's rl_end are
+ * same.
+ */
+static VALUE
+readline_s_get_line_buffer(VALUE self)
+{
+    rb_secure(4);
+    if (rl_line_buffer == NULL)
+	return Qnil;
+    return rb_locale_str_new_cstr(rl_line_buffer);
+}
+#else
+#define readline_s_get_line_buffer rb_f_notimplement
+#endif
+
+#ifdef HAVE_RL_POINT
+/*
+ * call-seq:
+ *   Readline.point -> int
+ *
+ * Returns the index of the current cursor position in
+ * +Readline.line_buffer+.
+ *
+ * The index in +Readline.line_buffer+ which matches the start of
+ * input-string passed to completion_proc is computed by subtracting
+ * the length of input-string from +Readline.point+.
+ *
+ *   start = (the length of input-string) - Readline.point
+ */
+static VALUE
+readline_s_get_point(VALUE self)
+{
+    rb_secure(4);
+    return INT2NUM(rl_point);
+}
+#else
+#define readline_s_get_point rb_f_notimplement
+#endif
+
 static char **
 readline_attempted_completion_function(const char *text, int start, int end)
 {
     VALUE proc, ary, temp;
     char **result;
     int case_fold;
-    int i, matches;
+    long i, matches;
+    rb_encoding *enc;
+    VALUE encobj;
 
     proc = rb_attr_get(mReadline, completion_proc);
     if (NIL_P(proc))
 	return NULL;
+#ifdef HAVE_RL_COMPLETION_APPEND_CHARACTER
+    rl_completion_append_character = readline_completion_append_character;
+#endif
 #ifdef HAVE_RL_ATTEMPTED_COMPLETION_OVER
     rl_attempted_completion_over = 1;
 #endif
     case_fold = RTEST(rb_attr_get(mReadline, completion_case_fold));
     ary = rb_funcall(proc, rb_intern("call"), 1, rb_locale_str_new_cstr(text));
-    if (TYPE(ary) != T_ARRAY)
+    if (!RB_TYPE_P(ary, T_ARRAY))
 	ary = rb_Array(ary);
     matches = RARRAY_LEN(ary);
-    if (matches == 0)
-	return NULL;
-    result = ALLOC_N(char *, matches + 2);
+    if (matches == 0) return NULL;
+    result = (char**)malloc((matches + 2)*sizeof(char*));
+    if (result == NULL) rb_raise(rb_eNoMemError, "failed to allocate memory");
+    enc = rb_locale_encoding();
+    encobj = rb_enc_from_encoding(enc);
     for (i = 0; i < matches; i++) {
 	temp = rb_obj_as_string(RARRAY_PTR(ary)[i]);
-	result[i + 1] = ALLOC_N(char, RSTRING_LEN(temp) + 1);
+	StringValueCStr(temp);	/* must be NUL-terminated */
+	rb_enc_check(encobj, temp);
+	result[i + 1] = (char*)malloc(RSTRING_LEN(temp) + 1);
+	if (result[i + 1]  == NULL) rb_memerror();
 	strcpy(result[i + 1], RSTRING_PTR(temp));
     }
     result[matches + 1] = NULL;
@@ -403,28 +691,27 @@ readline_attempted_completion_function(const char *text, int start, int end)
         result[0] = strdup(result[1]);
     }
     else {
-	register int i = 1;
-	int low = 100000;
+	const char *result1 = result[1];
+	long low = strlen(result1);
 
-	while (i < matches) {
-	    register int c1, c2, si;
+	for (i = 1; i < matches; ++i) {
+	    register int c1, c2;
+	    long i1, i2, l2;
+	    int n1, n2;
+	    const char *p2 = result[i + 1];
 
-	    if (case_fold) {
-		for (si = 0;
-		     (c1 = TOLOWER(result[i][si])) &&
-			 (c2 = TOLOWER(result[i + 1][si]));
-		     si++)
-		    if (c1 != c2) break;
-	    } else {
-		for (si = 0;
-		     (c1 = result[i][si]) &&
-			 (c2 = result[i + 1][si]);
-		     si++)
-		    if (c1 != c2) break;
+	    l2 = strlen(p2);
+	    for (i1 = i2 = 0; i1 < low && i2 < l2; i1 += n1, i2 += n2) {
+		c1 = rb_enc_codepoint_len(result1 + i1, result1 + low, &n1, enc);
+		c2 = rb_enc_codepoint_len(p2 + i2, p2 + l2, &n2, enc);
+		if (case_fold) {
+		    c1 = rb_tolower(c1);
+		    c2 = rb_tolower(c2);
+		}
+		if (c1 != c2) break;
 	    }
 
-	    if (low > si) low = si;
-	    i++;
+	    low = i1;
 	}
 	result[0] = ALLOC_N(char, low + 1);
 	strncpy(result[0], result[1], low);
@@ -434,6 +721,61 @@ readline_attempted_completion_function(const char *text, int start, int end)
     return result;
 }
 
+#ifdef HAVE_RL_SET_SCREEN_SIZE
+/*
+ * call-seq:
+ *   Readline.set_screen_size(rows, columns) -> self
+ *
+ * Set terminal size to +rows+ and +columns+.
+ *
+ * See GNU Readline's rl_set_screen_size function.
+ *
+ * Raises NotImplementedError if the using readline library does not support.
+ *
+ * Raises SecurityError exception if $SAFE is 4.
+ */
+static VALUE
+readline_s_set_screen_size(VALUE self, VALUE rows, VALUE columns)
+{
+    rb_secure(4);
+    rl_set_screen_size(NUM2INT(rows), NUM2INT(columns));
+    return self;
+}
+#else
+#define readline_s_set_screen_size rb_f_notimplement
+#endif
+
+#ifdef HAVE_RL_GET_SCREEN_SIZE
+/*
+ * call-seq:
+ *   Readline.get_screen_size -> [rows, columns]
+ *
+ * Returns the terminal's rows and columns.
+ *
+ * See GNU Readline's rl_get_screen_size function.
+ *
+ * Raises NotImplementedError if the using readline library does not support.
+ *
+ * Raises SecurityError exception if $SAFE is 4.
+ */
+static VALUE
+readline_s_get_screen_size(VALUE self)
+{
+    int rows, columns;
+    VALUE res;
+
+    rb_secure(4);
+    rl_get_screen_size(&rows, &columns);
+    res = rb_ary_new();
+    rb_ary_push(res, INT2NUM(rows));
+    rb_ary_push(res, INT2NUM(columns));
+    return res;
+}
+#else
+#define readline_s_get_screen_size rb_f_notimplement
+#endif
+
+#ifdef HAVE_RL_VI_EDITING_MODE
 /*
  * call-seq:
  *   Readline.vi_editing_mode -> nil
@@ -448,16 +790,15 @@ readline_attempted_completion_function(const char *text, int start, int end)
 static VALUE
 readline_s_vi_editing_mode(VALUE self)
 {
-#ifdef HAVE_RL_VI_EDITING_MODE
     rb_secure(4);
     rl_vi_editing_mode(1,0);
     return Qnil;
-#else
-    rb_notimplement();
-    return Qnil; /* not reached */
-#endif /* HAVE_RL_VI_EDITING_MODE */
 }
+#else
+#define readline_s_vi_editing_mode rb_f_notimplement
+#endif
 
+#ifdef HAVE_RL_EDITING_MODE
 /*
  * call-seq:
  *   Readline.vi_editing_mode? -> bool
@@ -471,15 +812,14 @@ readline_s_vi_editing_mode(VALUE self)
 static VALUE
 readline_s_vi_editing_mode_p(VALUE self)
 {
-#ifdef HAVE_RL_EDITING_MODE
     rb_secure(4);
     return rl_editing_mode == 0 ? Qtrue : Qfalse;
-#else
-    rb_notimplement();
-    return Qnil; /* not reached */
-#endif /* HAVE_RL_EDITING_MODE */
 }
+#else
+#define readline_s_vi_editing_mode_p rb_f_notimplement
+#endif
 
+#ifdef HAVE_RL_EMACS_EDITING_MODE
 /*
  * call-seq:
  *   Readline.emacs_editing_mode -> nil
@@ -494,16 +834,15 @@ readline_s_vi_editing_mode_p(VALUE self)
 static VALUE
 readline_s_emacs_editing_mode(VALUE self)
 {
-#ifdef HAVE_RL_EMACS_EDITING_MODE
     rb_secure(4);
     rl_emacs_editing_mode(1,0);
     return Qnil;
-#else
-    rb_notimplement();
-    return Qnil; /* not reached */
-#endif /* HAVE_RL_EMACS_EDITING_MODE */
 }
+#else
+#define readline_s_emacs_editing_mode rb_f_notimplement
+#endif
 
+#ifdef  HAVE_RL_EDITING_MODE
 /*
  * call-seq:
  *   Readline.emacs_editing_mode? -> bool
@@ -517,15 +856,14 @@ readline_s_emacs_editing_mode(VALUE self)
 static VALUE
 readline_s_emacs_editing_mode_p(VALUE self)
 {
-#ifdef  HAVE_RL_EDITING_MODE
     rb_secure(4);
     return rl_editing_mode == 1 ? Qtrue : Qfalse;
-#else
-    rb_notimplement();
-    return Qnil; /* not reached */
-#endif /* HAVE_RL_EDITING_MODE */
 }
+#else
+#define readline_s_emacs_editing_mode_p rb_f_notimplement
+#endif
 
+#ifdef HAVE_RL_COMPLETION_APPEND_CHARACTER
 /*
  * call-seq:
  *   Readline.completion_append_character = char
@@ -536,20 +874,20 @@ readline_s_emacs_editing_mode_p(VALUE self)
  *
  * For example:
  *   require "readline"
- *   
+ *
  *   Readline.readline("> ", true)
  *   Readline.completion_append_character = " "
  *
  * Result:
  *   >
  *   Input "/var/li".
- *   
+ *
  *   > /var/li
  *   Press TAB key.
- *   
+ *
  *   > /var/lib
  *   Completes "b" and appends " ". So, you can continuously input "/usr".
- *   
+ *
  *   > /var/lib /usr
  *
  * NOTE: Only one character can be specified. When "string" is
@@ -567,7 +905,6 @@ readline_s_emacs_editing_mode_p(VALUE self)
 static VALUE
 readline_s_set_completion_append_character(VALUE self, VALUE str)
 {
-#ifdef HAVE_RL_COMPLETION_APPEND_CHARACTER
     rb_secure(4);
     if (NIL_P(str)) {
 	rl_completion_append_character = '\0';
@@ -581,12 +918,12 @@ readline_s_set_completion_append_character(VALUE self, VALUE str)
 	}
     }
     return self;
-#else
-    rb_notimplement();
-    return Qnil; /* not reached */
-#endif /* HAVE_RL_COMPLETION_APPEND_CHARACTER */
 }
+#else
+#define readline_s_set_completion_append_character rb_f_notimplement
+#endif
 
+#ifdef HAVE_RL_COMPLETION_APPEND_CHARACTER
 /*
  * call-seq:
  *   Readline.completion_append_character -> char
@@ -601,7 +938,6 @@ readline_s_set_completion_append_character(VALUE self, VALUE str)
 static VALUE
 readline_s_get_completion_append_character(VALUE self)
 {
-#ifdef HAVE_RL_COMPLETION_APPEND_CHARACTER
     char buf[1];
 
     rb_secure(4);
@@ -610,12 +946,12 @@ readline_s_get_completion_append_character(VALUE self)
 
     buf[0] = (char) rl_completion_append_character;
     return rb_locale_str_new(buf, 1);
-#else
-    rb_notimplement();
-    return Qnil; /* not reached */
-#endif /* HAVE_RL_COMPLETION_APPEND_CHARACTER */
 }
+#else
+#define readline_s_get_completion_append_character rb_f_notimplement
+#endif
 
+#ifdef HAVE_RL_BASIC_WORD_BREAK_CHARACTERS
 /*
  * call-seq:
  *   Readline.basic_word_break_characters = string
@@ -631,7 +967,6 @@ readline_s_get_completion_append_character(VALUE self)
 static VALUE
 readline_s_set_basic_word_break_characters(VALUE self, VALUE str)
 {
-#ifdef HAVE_RL_BASIC_WORD_BREAK_CHARACTERS
     static char *basic_word_break_characters = NULL;
 
     rb_secure(4);
@@ -648,19 +983,19 @@ readline_s_set_basic_word_break_characters(VALUE self, VALUE str)
     basic_word_break_characters[RSTRING_LEN(str)] = '\0';
     rl_basic_word_break_characters = basic_word_break_characters;
     return self;
-#else
-    rb_notimplement();
-    return Qnil; /* not reached */
-#endif /* HAVE_RL_BASIC_WORD_BREAK_CHARACTERS */
 }
+#else
+#define readline_s_set_basic_word_break_characters rb_f_notimplement
+#endif
 
+#ifdef HAVE_RL_BASIC_WORD_BREAK_CHARACTERS
 /*
  * call-seq:
  *   Readline.basic_word_break_characters -> string
  *
  * Gets the basic list of characters that signal a break between words
  * for the completer routine.
- * 
+ *
  * Raises NotImplementedError if the using readline library does not support.
  *
  * Raises SecurityError exception if $SAFE is 4.
@@ -668,17 +1003,16 @@ readline_s_set_basic_word_break_characters(VALUE self, VALUE str)
 static VALUE
 readline_s_get_basic_word_break_characters(VALUE self, VALUE str)
 {
-#ifdef HAVE_RL_BASIC_WORD_BREAK_CHARACTERS
     rb_secure(4);
     if (rl_basic_word_break_characters == NULL)
 	return Qnil;
     return rb_locale_str_new_cstr(rl_basic_word_break_characters);
-#else
-    rb_notimplement();
-    return Qnil; /* not reached */
-#endif /* HAVE_RL_BASIC_WORD_BREAK_CHARACTERS */
 }
+#else
+#define readline_s_get_basic_word_break_characters rb_f_notimplement
+#endif
 
+#ifdef HAVE_RL_COMPLETER_WORD_BREAK_CHARACTERS
 /*
  * call-seq:
  *   Readline.completer_word_break_characters = string
@@ -694,7 +1028,6 @@ readline_s_get_basic_word_break_characters(VALUE self, VALUE str)
 static VALUE
 readline_s_set_completer_word_break_characters(VALUE self, VALUE str)
 {
-#ifdef HAVE_RL_COMPLETER_WORD_BREAK_CHARACTERS
     static char *completer_word_break_characters = NULL;
 
     rb_secure(4);
@@ -711,19 +1044,19 @@ readline_s_set_completer_word_break_characters(VALUE self, VALUE str)
     completer_word_break_characters[RSTRING_LEN(str)] = '\0';
     rl_completer_word_break_characters = completer_word_break_characters;
     return self;
-#else
-    rb_notimplement();
-    return Qnil; /* not reached */
-#endif /* HAVE_RL_COMPLETER_WORD_BREAK_CHARACTERS */
 }
+#else
+#define readline_s_set_completer_word_break_characters rb_f_notimplement
+#endif
 
+#ifdef HAVE_RL_COMPLETER_WORD_BREAK_CHARACTERS
 /*
  * call-seq:
  *   Readline.completer_word_break_characters -> string
  *
  * Gets the basic list of characters that signal a break between words
  * for rl_complete_internal().
- * 
+ *
  * Raises NotImplementedError if the using readline library does not support.
  *
  * Raises SecurityError exception if $SAFE is 4.
@@ -731,17 +1064,16 @@ readline_s_set_completer_word_break_characters(VALUE self, VALUE str)
 static VALUE
 readline_s_get_completer_word_break_characters(VALUE self, VALUE str)
 {
-#ifdef HAVE_RL_COMPLETER_WORD_BREAK_CHARACTERS
     rb_secure(4);
     if (rl_completer_word_break_characters == NULL)
 	return Qnil;
     return rb_locale_str_new_cstr(rl_completer_word_break_characters);
-#else
-    rb_notimplement();
-    return Qnil; /* not reached */
-#endif /* HAVE_RL_COMPLETER_WORD_BREAK_CHARACTERS */
 }
+#else
+#define readline_s_get_completer_word_break_characters rb_f_notimplement
+#endif
 
+#ifdef HAVE_RL_BASIC_QUOTE_CHARACTERS
 /*
  * call-seq:
  *   Readline.basic_quote_characters = string
@@ -755,7 +1087,6 @@ readline_s_get_completer_word_break_characters(VALUE self, VALUE str)
 static VALUE
 readline_s_set_basic_quote_characters(VALUE self, VALUE str)
 {
-#ifdef HAVE_RL_BASIC_QUOTE_CHARACTERS
     static char *basic_quote_characters = NULL;
 
     rb_secure(4);
@@ -773,12 +1104,12 @@ readline_s_set_basic_quote_characters(VALUE self, VALUE str)
     rl_basic_quote_characters = basic_quote_characters;
 
     return self;
-#else
-    rb_notimplement();
-    return Qnil; /* not reached */
-#endif /* HAVE_RL_BASIC_QUOTE_CHARACTERS */
 }
+#else
+#define readline_s_set_basic_quote_characters rb_f_notimplement
+#endif
 
+#ifdef HAVE_RL_BASIC_QUOTE_CHARACTERS
 /*
  * call-seq:
  *   Readline.basic_quote_characters -> string
@@ -792,17 +1123,16 @@ readline_s_set_basic_quote_characters(VALUE self, VALUE str)
 static VALUE
 readline_s_get_basic_quote_characters(VALUE self, VALUE str)
 {
-#ifdef HAVE_RL_BASIC_QUOTE_CHARACTERS
     rb_secure(4);
     if (rl_basic_quote_characters == NULL)
 	return Qnil;
     return rb_locale_str_new_cstr(rl_basic_quote_characters);
-#else
-    rb_notimplement();
-    return Qnil; /* not reached */
-#endif /* HAVE_RL_BASIC_QUOTE_CHARACTERS */
 }
+#else
+#define readline_s_get_basic_quote_characters rb_f_notimplement
+#endif
 
+#ifdef HAVE_RL_COMPLETER_QUOTE_CHARACTERS
 /*
  * call-seq:
  *   Readline.completer_quote_characters = string
@@ -819,7 +1149,6 @@ readline_s_get_basic_quote_characters(VALUE self, VALUE str)
 static VALUE
 readline_s_set_completer_quote_characters(VALUE self, VALUE str)
 {
-#ifdef HAVE_RL_COMPLETER_QUOTE_CHARACTERS
     static char *completer_quote_characters = NULL;
 
     rb_secure(4);
@@ -836,12 +1165,12 @@ readline_s_set_completer_quote_characters(VALUE self, VALUE str)
     rl_completer_quote_characters = completer_quote_characters;
 
     return self;
-#else
-    rb_notimplement();
-    return Qnil; /* not reached */
-#endif /* HAVE_RL_COMPLETER_QUOTE_CHARACTERS */
 }
+#else
+#define readline_s_set_completer_quote_characters rb_f_notimplement
+#endif
 
+#ifdef HAVE_RL_COMPLETER_QUOTE_CHARACTERS
 /*
  * call-seq:
  *   Readline.completer_quote_characters -> string
@@ -856,17 +1185,16 @@ readline_s_set_completer_quote_characters(VALUE self, VALUE str)
 static VALUE
 readline_s_get_completer_quote_characters(VALUE self, VALUE str)
 {
-#ifdef HAVE_RL_COMPLETER_QUOTE_CHARACTERS
     rb_secure(4);
     if (rl_completer_quote_characters == NULL)
 	return Qnil;
     return rb_locale_str_new_cstr(rl_completer_quote_characters);
-#else
-    rb_notimplement();
-    return Qnil; /* not reached */
-#endif /* HAVE_RL_COMPLETER_QUOTE_CHARACTERS */
 }
+#else
+#define readline_s_get_completer_quote_characters rb_f_notimplement
+#endif
 
+#ifdef HAVE_RL_FILENAME_QUOTE_CHARACTERS
 /*
  * call-seq:
  *   Readline.filename_quote_characters = string
@@ -881,7 +1209,6 @@ readline_s_get_completer_quote_characters(VALUE self, VALUE str)
 static VALUE
 readline_s_set_filename_quote_characters(VALUE self, VALUE str)
 {
-#ifdef HAVE_RL_FILENAME_QUOTE_CHARACTERS
     static char *filename_quote_characters = NULL;
 
     rb_secure(4);
@@ -898,12 +1225,12 @@ readline_s_set_filename_quote_characters(VALUE self, VALUE str)
     rl_filename_quote_characters = filename_quote_characters;
 
     return self;
-#else
-    rb_notimplement();
-    return Qnil; /* not reached */
-#endif /* HAVE_RL_FILENAME_QUOTE_CHARACTERS */
 }
+#else
+#define readline_s_set_filename_quote_characters rb_f_notimplement
+#endif
 
+#ifdef HAVE_RL_FILENAME_QUOTE_CHARACTERS
 /*
  * call-seq:
  *   Readline.filename_quote_characters -> string
@@ -918,16 +1245,34 @@ readline_s_set_filename_quote_characters(VALUE self, VALUE str)
 static VALUE
 readline_s_get_filename_quote_characters(VALUE self, VALUE str)
 {
-#ifdef HAVE_RL_FILENAME_QUOTE_CHARACTERS
     rb_secure(4);
     if (rl_filename_quote_characters == NULL)
 	return Qnil;
     return rb_locale_str_new_cstr(rl_filename_quote_characters);
-#else
-    rb_notimplement();
-    return Qnil; /* not reached */
-#endif /* HAVE_RL_FILENAME_QUOTE_CHARACTERS */
 }
+#else
+#define readline_s_get_filename_quote_characters rb_f_notimplement
+#endif
+
+#ifdef HAVE_RL_REFRESH_LINE
+/*
+ * call-seq:
+ *   Readline.refresh_line -> nil
+ *
+ * Clear the current input line.
+ *
+ * Raises SecurityError exception if $SAFE is 4.
+ */
+static VALUE
+readline_s_refresh_line(VALUE self)
+{
+    rb_secure(4);
+    rl_refresh_line(0, 0);
+    return Qnil;
+}
+#else
+#define readline_s_refresh_line rb_f_notimplement
+#endif
 
 static VALUE
 hist_to_s(VALUE self)
@@ -967,10 +1312,10 @@ hist_get(VALUE self, VALUE index)
     return rb_locale_str_new_cstr(entry->line);
 }
 
+#ifdef HAVE_REPLACE_HISTORY_ENTRY
 static VALUE
 hist_set(VALUE self, VALUE index, VALUE str)
 {
-#ifdef HAVE_REPLACE_HISTORY_ENTRY
     HIST_ENTRY *entry = NULL;
     int i;
 
@@ -981,17 +1326,16 @@ hist_set(VALUE self, VALUE index, VALUE str)
         i += history_length;
     }
     if (i >= 0) {
-	entry = replace_history_entry(i, RSTRING_PTR(str), NULL);
+	entry = replace_history_entry(history_replace_offset_func(i), RSTRING_PTR(str), NULL);
     }
     if (entry == NULL) {
 	rb_raise(rb_eIndexError, "invalid index");
     }
     return str;
-#else
-    rb_notimplement();
-    return Qnil; /* not reached */
-#endif
 }
+#else
+#define hist_set rb_f_notimplement
+#endif
 
 static VALUE
 hist_push(VALUE self, VALUE str)
@@ -1006,7 +1350,7 @@ static VALUE
 hist_push_method(int argc, VALUE *argv, VALUE self)
 {
     VALUE str;
-    
+
     rb_secure(4);
     while (argc--) {
 	str = *argv++;
@@ -1107,18 +1451,17 @@ hist_delete_at(VALUE self, VALUE index)
     return rb_remove_history(i);
 }
 
+#ifdef HAVE_CLEAR_HISTORY
 static VALUE
 hist_clear(VALUE self)
 {
-#ifdef HAVE_CLEAR_HISTORY
     rb_secure(4);
     clear_history();
     return self;
-#else
-    rb_notimplement();
-    return Qnil; /* not reached */
-#endif
 }
+#else
+#define hist_clear rb_f_notimplement
+#endif
 
 static VALUE
 filename_completion_proc_call(VALUE self, VALUE str)
@@ -1178,6 +1521,16 @@ Init_readline()
     /* Allow conditional parsing of the ~/.inputrc file. */
     rl_readline_name = (char *)"Ruby";
 
+#if defined HAVE_RL_GETC_FUNCTION
+    /* libedit check rl_getc_function only when rl_initialize() is called, */
+    /* and using_history() call rl_initialize(). */
+    /* This assignment should be placed before using_history() */
+    rl_getc_function = readline_getc;
+    id_getbyte = rb_intern_const("getbyte");
+#elif defined HAVE_RL_EVENT_HOOK
+    rl_event_hook = readline_event;
+#endif
+
     using_history();
 
     completion_proc = rb_intern(COMPLETION_PROC);
@@ -1198,6 +1551,14 @@ Init_readline()
 			       readline_s_set_completion_case_fold, 1);
     rb_define_singleton_method(mReadline, "completion_case_fold",
 			       readline_s_get_completion_case_fold, 0);
+    rb_define_singleton_method(mReadline, "line_buffer",
+			       readline_s_get_line_buffer, 0);
+    rb_define_singleton_method(mReadline, "point",
+			       readline_s_get_point, 0);
+    rb_define_singleton_method(mReadline, "set_screen_size",
+			       readline_s_set_screen_size, 2);
+    rb_define_singleton_method(mReadline, "get_screen_size",
+			       readline_s_get_screen_size, 0);
     rb_define_singleton_method(mReadline, "vi_editing_mode",
 			       readline_s_vi_editing_mode, 0);
     rb_define_singleton_method(mReadline, "vi_editing_mode?",
@@ -1230,6 +1591,13 @@ Init_readline()
 			       readline_s_set_filename_quote_characters, 1);
     rb_define_singleton_method(mReadline, "filename_quote_characters",
 			       readline_s_get_filename_quote_characters, 0);
+    rb_define_singleton_method(mReadline, "refresh_line",
+			       readline_s_refresh_line, 0);
+
+#if USE_INSERT_IGNORE_ESCAPE
+    CONST_ID(id_orig_prompt, "orig_prompt");
+    CONST_ID(id_last_prompt, "last_prompt");
+#endif
 
     history = rb_obj_alloc(rb_cObject);
     rb_extend_object(history, rb_mEnumerable);
@@ -1273,16 +1641,22 @@ Init_readline()
      */
     rb_define_const(mReadline, "USERNAME_COMPLETION_PROC", ucomp);
     history_get_offset_func = history_get_offset_history_base;
+    history_replace_offset_func = history_get_offset_0;
 #if defined HAVE_RL_LIBRARY_VERSION
     version = rb_str_new_cstr(rl_library_version);
 #if defined HAVE_CLEAR_HISTORY || defined HAVE_REMOVE_HISTORY
-    if (strncmp(rl_library_version, EDIT_LINE_LIBRARY_VERSION, 
+    if (strncmp(rl_library_version, EDIT_LINE_LIBRARY_VERSION,
 		strlen(EDIT_LINE_LIBRARY_VERSION)) == 0) {
 	add_history("1");
 	if (history_get(history_get_offset_func(0)) == NULL) {
 	    history_get_offset_func = history_get_offset_0;
 	}
-#if !defined HAVE_CLEAR_HISTORY
+#ifdef HAVE_REPLACE_HISTORY_ENTRY
+	if (replace_history_entry(0, "a", NULL) == NULL) {
+	    history_replace_offset_func = history_get_offset_history_base;
+	}
+#endif
+#ifdef HAVE_CLEAR_HISTORY
 	clear_history();
 #else
 	{
@@ -1302,10 +1676,13 @@ Init_readline()
     rb_define_const(mReadline, "VERSION", version);
 
     rl_attempted_completion_function = readline_attempted_completion_function;
-#ifdef HAVE_RL_EVENT_HOOK
-    rl_event_hook = readline_event;
-#endif
+#ifdef HAVE_RL_CATCH_SIGNALS
+    rl_catch_signals = 0;
+#endif 
+
 #ifdef HAVE_RL_CLEAR_SIGNALS
     rl_clear_signals();
 #endif
+
+    readline_s_set_input(mReadline, rb_stdin);
 }
