@@ -1,13 +1,12 @@
 /*
  * node.c
  *
- * $Author: shyouhei $
- * $Date: 2007-02-13 08:01:19 +0900 (Tue, 13 Feb 2007) $
+ * $Author: nobu $
  *
  * Copyright (C) 2003 why the lucky stiff
  */
 
-#include "ruby.h"
+#include "ruby/ruby.h"
 #include "syck.h"
 
 /*
@@ -46,7 +45,7 @@ syck_free_node( SyckNode *n )
 }
 
 SyckNode *
-syck_alloc_map()
+syck_alloc_map(void)
 {
     SyckNode *n;
     struct SyckMap *m;
@@ -60,12 +59,12 @@ syck_alloc_map()
 
     n = syck_alloc_node( syck_map_kind );
     n->data.pairs = m;
-    
+
     return n;
 }
 
 SyckNode *
-syck_alloc_seq()
+syck_alloc_seq(void)
 {
     SyckNode *n;
     struct SyckSeq *s;
@@ -83,7 +82,7 @@ syck_alloc_seq()
 }
 
 SyckNode *
-syck_alloc_str()
+syck_alloc_str(void)
 {
     SyckNode *n;
     struct SyckStr *s;
@@ -95,18 +94,18 @@ syck_alloc_str()
 
     n = syck_alloc_node( syck_str_kind );
     n->data.str = s;
-    
+
     return n;
 }
 
 SyckNode *
-syck_new_str( char *str, enum scalar_style style )
+syck_new_str( const char *str, enum scalar_style style )
 {
     return syck_new_str2( str, strlen( str ), style );
 }
 
 SyckNode *
-syck_new_str2( char *str, long len, enum scalar_style style )
+syck_new_str2( const char *str, long len, enum scalar_style style )
 {
     SyckNode *n;
 
@@ -129,7 +128,7 @@ syck_replace_str( SyckNode *n, char *str, enum scalar_style style )
 void
 syck_replace_str2( SyckNode *n, char *str, long len, enum scalar_style style )
 {
-    if ( n->data.str != NULL ) 
+    if ( n->data.str->ptr != NULL )
     {
         S_FREE( n->data.str->ptr );
         n->data.str->ptr = NULL;
@@ -202,7 +201,7 @@ syck_map_add( SyckNode *map, SYMID key, SYMID value )
 
     ASSERT( map != NULL );
     ASSERT( map->data.pairs != NULL );
-    
+
     m = map->data.pairs;
     idx = m->idx;
     m->idx += 1;
@@ -227,7 +226,7 @@ syck_map_update( SyckNode *map1, SyckNode *map2 )
     m1 = map1->data.pairs;
     m2 = map2->data.pairs;
     if ( m2->idx < 1 ) return;
-        
+
     new_idx = m1->idx;
     new_idx += m2->idx;
     new_capa = m1->capa;
@@ -243,8 +242,8 @@ syck_map_update( SyckNode *map1, SyckNode *map2 )
     }
     for ( new_idx = 0; new_idx < m2->idx; m1->idx++, new_idx++ )
     {
-        m1->keys[m1->idx] = m2->keys[new_idx]; 
-        m1->values[m1->idx] = m2->values[new_idx]; 
+        m1->keys[m1->idx] = m2->keys[new_idx];
+        m1->values[m1->idx] = m2->values[new_idx];
     }
 }
 
@@ -325,7 +324,7 @@ syck_seq_add( SyckNode *arr, SYMID value )
 
     ASSERT( arr != NULL );
     ASSERT( arr->data.list != NULL );
-    
+
     s = arr->data.list;
     idx = s->idx;
     s->idx += 1;
@@ -375,7 +374,7 @@ syck_free_members( SyckNode *n )
     switch ( n->kind  )
     {
         case syck_str_kind:
-            if ( n->data.str != NULL ) 
+            if ( n->data.str != NULL )
             {
                 S_FREE( n->data.str->ptr );
                 n->data.str->ptr = NULL;
