@@ -6,6 +6,12 @@
 
 #include "ruby.h"
 #include "stubs.h"
+
+#if !defined(RSTRING_PTR)
+#define RSTRING_PTR(s) (RSTRING(s)->ptr)
+#define RSTRING_LEN(s) (RSTRING(s)->len)
+#endif
+
 #include <tcl.h>
 #include <tk.h>
 
@@ -40,7 +46,7 @@ _nativethread_consistency_check(ip)
         return;
     }
 
-    /* If the variable "tcl_platform(threaded)" exists,  
+    /* If the variable "tcl_platform(threaded)" exists,
        then the Tcl interpreter was compiled with threads enabled. */
     if (Tcl_GetVar2(ip, "tcl_platform", "threaded", TCL_GLOBAL_ONLY) != (char*)NULL) {
 #ifdef HAVE_NATIVETHREAD
@@ -66,15 +72,19 @@ _nativethread_consistency_check(ip)
 #if defined USE_TCL_STUBS && defined USE_TK_STUBS
 
 #if defined _WIN32 || defined __CYGWIN__
-# include "util.h"
+#  ifdef HAVE_RUBY_RUBY_H
+#    include "ruby/util.h"
+#  else
+#    include "util.h"
+#  endif
 # include <windows.h>
   typedef HINSTANCE DL_HANDLE;
 # define DL_OPEN LoadLibrary
 # define DL_SYM GetProcAddress
 # define TCL_INDEX 4
 # define TK_INDEX 3
-# define TCL_NAME "tcl89%s"
-# define TK_NAME "tk89%s"
+# define TCL_NAME "tcl89"
+# define TK_NAME "tk89"
 # undef DLEXT
 # define DLEXT ".dll"
 #elif defined HAVE_DLOPEN
@@ -84,8 +94,12 @@ _nativethread_consistency_check(ip)
 # define DL_SYM dlsym
 # define TCL_INDEX 8
 # define TK_INDEX 7
-# define TCL_NAME "libtcl8.9%s"
-# define TK_NAME "libtk8.9%s"
+# define TCL_NAME "libtcl8.9"
+# define TK_NAME "libtk8.9"
+# if defined(__APPLE__) && defined(__MACH__)   /* Mac OS X */
+#  undef DLEXT
+#  define DLEXT ".dylib"
+# endif
 #endif
 
 static DL_HANDLE tcl_dll = (DL_HANDLE)0;
@@ -102,7 +116,6 @@ ruby_open_tcl_dll(appname)
     void (*p_Tcl_FindExecutable)(const char *);
     int n;
     char *ruby_tcl_dll = 0;
-    char tcl_name[20];
 
     if (tcl_dll) return TCLTK_STUBS_OK;
 
@@ -113,7 +126,7 @@ ruby_open_tcl_dll(appname)
     if (ruby_tcl_dll) {
         tcl_dll = (DL_HANDLE)DL_OPEN(ruby_tcl_dll);
     } else {
-        snprintf(tcl_name, sizeof tcl_name, TCL_NAME, DLEXT);
+	char tcl_name[] = TCL_NAME DLEXT;
         /* examine from 8.9 to 8.1 */
         for (n = '9'; n > '0'; n--) {
             tcl_name[TCL_INDEX] = n;
@@ -148,7 +161,6 @@ ruby_open_tk_dll()
 {
     int n;
     char *ruby_tk_dll = 0;
-    char tk_name[20];
 
     if (!tcl_dll) {
         /* int ret = ruby_open_tcl_dll(RSTRING_PTR(rb_argv0)); */
@@ -162,7 +174,7 @@ ruby_open_tk_dll()
     if (ruby_tk_dll) {
         tk_dll = (DL_HANDLE)DL_OPEN(ruby_tk_dll);
     } else {
-        snprintf(tk_name, sizeof tk_name, TK_NAME, DLEXT);
+	char tk_name[] = TK_NAME DLEXT;
         /* examine from 8.9 to 8.1 */
         for (n = '9'; n > '0'; n--) {
             tk_name[TK_INDEX] = n;
@@ -189,13 +201,13 @@ ruby_open_tcltk_dll(appname)
     return( ruby_open_tcl_dll(appname) || ruby_open_tk_dll() );
 }
 
-int 
+int
 tcl_stubs_init_p()
 {
     return(tclStubsPtr != (TclStubs*)NULL);
 }
 
-int 
+int
 tk_stubs_init_p()
 {
     return(tkStubsPtr != (TkStubs*)NULL);
@@ -240,14 +252,14 @@ ruby_tcl_create_ip_and_stubs_init(st)
             }
         }
 
-        p_Tcl_CreateInterp 
+        p_Tcl_CreateInterp
             = (Tcl_Interp *(*)())DL_SYM(tcl_dll, "Tcl_CreateInterp");
         if (!p_Tcl_CreateInterp) {
             if (st) *st = NO_CreateInterp;
             return (Tcl_Interp*)NULL;
         }
 
-        p_Tcl_DeleteInterp 
+        p_Tcl_DeleteInterp
             = (Tcl_Interp *(*)())DL_SYM(tcl_dll, "Tcl_DeleteInterp");
         if (!p_Tcl_DeleteInterp) {
             if (st) *st = NO_DeleteInterp;
@@ -315,10 +327,26 @@ ruby_tk_stubs_init(tcl_ip)
         if (!p_Tk_Init)
             return NO_Tk_Init;
 
+#if defined USE_TK_STUBS && defined TK_FRAMEWORK && defined(__APPLE__) && defined(__MACH__)
+	/*
+	  FIX ME : dirty hack for Mac OS X frameworks.
+	  With stubs, fails to find Resource/Script directory of Tk.framework.
+	  So, teach it to a Tcl interpreter by an environment variable.
+	  e.g. when $tcl_library ==
+	               /Library/Frameworks/Tcl.framwwork/8.5/Resources/Scripts
+		   ==> /Library/Frameworks/Tk.framwwork/8.5/Resources/Scripts
+	*/
+	if (Tcl_Eval(tcl_ip,
+		     "if {[array get env TK_LIBRARY] == {}} { set env(TK_LIBRARY) [regsub -all -nocase {(t)cl} $tcl_library  {\\1k}] }"
+		     ) != TCL_OK) {
+	  return FAIL_Tk_Init;
+	}
+#endif
+
         if ((*p_Tk_Init)(tcl_ip) == TCL_ERROR)
             return FAIL_Tk_Init;
 
-        if (!Tk_InitStubs(tcl_ip, "8.1", 0))
+        if (!Tk_InitStubs(tcl_ip, (char *)"8.1", 0))
             return FAIL_Tk_InitStubs;
 
 #ifdef __MACOS__
@@ -357,7 +385,7 @@ ruby_tk_stubs_safeinit(tcl_ip)
         if ((*p_Tk_SafeInit)(tcl_ip) == TCL_ERROR)
             return FAIL_Tk_Init;
 
-        if (!Tk_InitStubs(tcl_ip, "8.1", 0))
+        if (!Tk_InitStubs(tcl_ip, (char *)"8.1", 0))
             return FAIL_Tk_InitStubs;
 
 #ifdef __MACOS__
@@ -438,7 +466,7 @@ ruby_open_tcl_dll(appname)
     return TCLTK_STUBS_OK;
 }
 
-int 
+int
 ruby_open_tk_dll()
 {
     if (!open_tcl_dll) {
@@ -449,7 +477,7 @@ ruby_open_tk_dll()
     return TCLTK_STUBS_OK;
 }
 
-int 
+int
 #ifdef HAVE_PROTOTYPES
 ruby_open_tcltk_dll(char *appname)
 #else
@@ -460,13 +488,13 @@ ruby_open_tcltk_dll(appname)
     return( ruby_open_tcl_dll(appname) || ruby_open_tk_dll() );
 }
 
-int 
+int
 tcl_stubs_init_p()
 {
     return 1;
 }
 
-int 
+int
 tk_stubs_init_p()
 {
     return call_tk_stubs_init;
@@ -499,13 +527,13 @@ ruby_tcl_create_ip_and_stubs_init(st)
     return tcl_ip;
 }
 
-int 
+int
 ruby_tcl_stubs_init()
 {
     return TCLTK_STUBS_OK;
 }
 
-int 
+int
 #ifdef HAVE_PROTOTYPES
 ruby_tk_stubs_init(Tcl_Interp *tcl_ip)
 #else
@@ -553,7 +581,7 @@ ruby_tk_stubs_safeinit(tcl_ip)
 #endif
 }
 
-int 
+int
 ruby_tcltk_stubs()
 {
     /* Tcl_FindExecutable(RSTRING_PTR(rb_argv0)); */
