@@ -25,6 +25,8 @@
 # +:sources+:: Sets Gem::sources
 # +:verbose+:: See #verbose
 
+require 'rbconfig'
+
 class Gem::ConfigFile
 
   DEFAULT_BACKTRACE = false
@@ -47,10 +49,29 @@ class Gem::ConfigFile
 
   system_config_path =
     begin
-      require 'etc.so'
+      require "etc"
       Etc.sysconfdir
-    rescue LoadError
-      '/etc'
+    rescue LoadError, NoMethodError
+      begin
+        # TODO: remove after we drop 1.8.7 and 1.9.1
+        require 'Win32API'
+
+        CSIDL_COMMON_APPDATA = 0x0023
+        path = 0.chr * 260
+        if RUBY_VERSION > '1.9' then
+          SHGetFolderPath = Win32API.new 'shell32', 'SHGetFolderPath', 'PLPLP',
+          'L', :stdcall
+          SHGetFolderPath.call nil, CSIDL_COMMON_APPDATA, nil, 1, path
+        else
+          SHGetFolderPath = Win32API.new 'shell32', 'SHGetFolderPath', 'LLLLP',
+          'L'
+          SHGetFolderPath.call 0, CSIDL_COMMON_APPDATA, 0, 1, path
+        end
+
+        path.strip
+      rescue LoadError
+        RbConfig::CONFIG["sysconfdir"] || "/etc"
+      end
     end
 
   SYSTEM_WIDE_CONFIG_FILE = File.join system_config_path, 'gemrc'
@@ -104,6 +125,11 @@ class Gem::ConfigFile
 
   attr_reader :rubygems_api_key
 
+  ##
+  # Hash of RubyGems.org and alternate API keys
+
+  attr_reader :api_keys
+
   ##
   # openssl verify mode value, used for remote https connection
 
@@ -179,7 +205,7 @@ class Gem::ConfigFile
     @ssl_verify_mode  = @hash[:ssl_verify_mode]  if @hash.key? :ssl_verify_mode
     @ssl_ca_cert      = @hash[:ssl_ca_cert]      if @hash.key? :ssl_ca_cert
 
-    load_rubygems_api_key
+    load_api_keys
 
     Gem.sources = @hash[:sources] if @hash.key? :sources
     handle_arguments arg_list
@@ -189,22 +215,28 @@ class Gem::ConfigFile
   # Location of RubyGems.org credentials
 
   def credentials_path
-    File.join(Gem.user_home, '.gem', 'credentials')
+    File.join Gem.user_home, '.gem', 'credentials'
   end
 
-  def load_rubygems_api_key
-    api_key_hash = File.exists?(credentials_path) ? load_file(credentials_path) : @hash
-
-    @rubygems_api_key = api_key_hash[:rubygems_api_key] if api_key_hash.key? :rubygems_api_key
+  def load_api_keys
+    @api_keys = if File.exist? credentials_path then
+                  load_file(credentials_path)
+                else
+                  @hash
+                end
+    if @api_keys.key? :rubygems_api_key then
+      @rubygems_api_key = @api_keys[:rubygems_api_key]
+      @api_keys[:rubygems] = @api_keys.delete :rubygems_api_key unless @api_keys.key? :rubygems
+    end
   end
 
   def rubygems_api_key=(api_key)
     config = load_file(credentials_path).merge(:rubygems_api_key => api_key)
 
-    dirname = File.dirname(credentials_path)
-    Dir.mkdir(dirname) unless File.exists?(dirname)
+    dirname = File.dirname credentials_path
+    Dir.mkdir(dirname) unless File.exist? dirname
 
-    require 'yaml'
+    Gem.load_yaml
 
     File.open(credentials_path, 'w') do |f|
       f.write config.to_yaml
@@ -214,9 +246,10 @@ class Gem::ConfigFile
   end
 
   def load_file(filename)
-    return {} unless filename and File.exists?(filename)
+    Gem.load_yaml
+
+    return {} unless filename and File.exist? filename
     begin
-      require 'yaml'
       YAML.load(File.read(filename))
     rescue ArgumentError
       warn "Failed to load #{config_file_name}"
@@ -311,7 +344,6 @@ class Gem::ConfigFile
 
   # Writes out this config file, replacing its source.
   def write
-    require 'yaml'
     open config_file_name, 'w' do |io|
       io.write to_yaml
     end
@@ -340,6 +372,4 @@ class Gem::ConfigFile
   protected
 
   attr_reader :hash
-
 end
-
