@@ -201,20 +201,14 @@ class TestArgf < Test::Unit::TestCase
     t = make_tempfile
 
     assert_in_out_err(["-", t.path], <<-INPUT) do |r, e|
-      ARGF.inplace_mode = '/\\\\'
+      ARGF.inplace_mode = '/\\\\:'
       while line = ARGF.gets
         puts line.chomp + '.new'
       end
     INPUT
-      if no_safe_rename
-        assert_equal([], e)
-        assert_equal([], r)
-        assert_equal("foo.new\nbar.new\nbaz.new\n", File.read(t.path))
-      else
-        assert_match(/Can't rename .* to .*: .*. skipping file/, e.first) #'
-        assert_equal([], r)
-        assert_equal("foo\nbar\nbaz\n", File.read(t.path))
-      end
+      assert_match(/Can't rename .* to .*: .*. skipping file/, e.first) #'
+      assert_equal([], r)
+      assert_equal("foo\nbar\nbaz\n", File.read(t.path))
     end
   end
 
@@ -647,12 +641,23 @@ class TestArgf < Test::Unit::TestCase
   end
 
   def test_binmode
+    bug5268 = '[ruby-core:39234]'
+    open(@t3.path, "wb") {|f| f.write "5\r\n6\r\n"}
     ruby('-e', "ARGF.binmode; STDOUT.binmode; puts ARGF.read", @t1.path, @t2.path, @t3.path) do |f|
       f.binmode
-      assert_equal("1\n2\n3\n4\n5\n6\n", f.read)
+      assert_equal("1\n2\n3\n4\n5\r\n6\r\n", f.read, bug5268)
     end
   end
 
+  def test_textmode
+    bug5268 = '[ruby-core:39234]'
+    open(@t3.path, "wb") {|f| f.write "5\r\n6\r\n"}
+    ruby('-e', "STDOUT.binmode; puts ARGF.read", @t1.path, @t2.path, @t3.path) do |f|
+      f.binmode
+      assert_equal("1\n2\n3\n4\n5\n6\n", f.read, bug5268)
+    end
+  end unless IO::BINARY.zero?
+
   def test_skip
     ruby('-e', <<-SRC, @t1.path, @t2.path, @t3.path) do |f|
       ARGF.skip
@@ -673,6 +678,16 @@ class TestArgf < Test::Unit::TestCase
     end
   end
 
+  def test_close_replace
+    ruby('-e', <<-SRC) do |f|
+      ARGF.close
+      ARGV.replace ['#{@t1.path}', '#{@t2.path}', '#{@t3.path}']
+      puts ARGF.read
+    SRC
+      assert_equal("1\n2\n3\n4\n5\n6\n", f.read)
+    end
+  end
+
   def test_closed
     ruby('-e', <<-SRC, @t1.path, @t2.path, @t3.path) do |f|
       3.times do
@@ -695,6 +710,32 @@ class TestArgf < Test::Unit::TestCase
     end
   end
 
+  def test_readlines_limit_0
+    bug4024 = '[ruby-dev:42538]'
+    t = make_tempfile
+    argf = ARGF.class.new(t.path)
+    begin
+      assert_raise(ArgumentError, bug4024) do
+        argf.readlines(0)
+      end
+    ensure
+      argf.close
+    end
+  end
+
+  def test_each_line_limit_0
+    bug4024 = '[ruby-dev:42538]'
+    t = make_tempfile
+    argf = ARGF.class.new(t.path)
+    begin
+      assert_raise(ArgumentError, bug4024) do
+        argf.each_line(0).next
+      end
+    ensure
+      argf.close
+    end
+  end
+
   def test_unreadable
     bug4274 = '[ruby-core:34446]'
     paths = (1..2).map do
@@ -710,4 +751,9 @@ class TestArgf < Test::Unit::TestCase
     end
     assert_nil(argf.gets, bug4274)
   end
+
+  def test_readlines_twice
+    bug5952 = '[ruby-dev:45160]'
+    assert_ruby_status(["-e", "2.times {STDIN.tty?; readlines}"], "", bug5952)
+  end
 end
