@@ -2,6 +2,8 @@ require 'test/unit'
 require_relative 'envutil'
 
 class TestFloat < Test::Unit::TestCase
+  include EnvUtil
+
   def test_float
     assert_equal(2, 2.6.floor)
     assert_equal(-3, (-2.6).floor)
@@ -86,19 +88,28 @@ class TestFloat < Test::Unit::TestCase
     assert_raise(ArgumentError){Float("-.")}
     assert_raise(ArgumentError){Float("1e")}
     assert_raise(ArgumentError){Float("1__1")}
+    assert_raise(ArgumentError){Float("1.")}
+    assert_raise(ArgumentError){Float("1.e+00")}
+    assert_raise(ArgumentError){Float("0x1.p+0")}
     # add expected behaviour here.
     assert_equal(10, Float("1_0"))
 
     assert_equal([ 0.0].pack('G'), [Float(" 0x0p+0").to_f].pack('G'))
     assert_equal([-0.0].pack('G'), [Float("-0x0p+0").to_f].pack('G'))
     assert_equal(255.0,     Float("0Xff"))
-    assert_equal(1.0,       Float("0X1.P+0"))
     assert_equal(1024.0,    Float("0x1p10"))
     assert_equal(1024.0,    Float("0x1p+10"))
     assert_equal(0.0009765625, Float("0x1p-10"))
     assert_equal(2.6881171418161356e+43, Float("0x1.3494a9b171bf5p+144"))
     assert_equal(-3.720075976020836e-44, Float("-0x1.a8c1f14e2af5dp-145"))
- 
+    assert_equal(31.0*2**1019, Float("0x0."+("0"*268)+"1fp2099"))
+    assert_equal(31.0*2**1019, Float("0x0."+("0"*600)+"1fp3427"))
+    assert_equal(-31.0*2**1019, Float("-0x0."+("0"*268)+"1fp2099"))
+    assert_equal(-31.0*2**1019, Float("-0x0."+("0"*600)+"1fp3427"))
+    assert_equal(31.0*2**-1027, Float("0x1f"+("0"*268)+".0p-2099"))
+    assert_equal(31.0*2**-1027, Float("0x1f"+("0"*600)+".0p-3427"))
+    assert_equal(-31.0*2**-1027, Float("-0x1f"+("0"*268)+".0p-2099"))
+    assert_equal(-31.0*2**-1027, Float("-0x1f"+("0"*600)+".0p-3427"))
   end
 
   def test_divmod
@@ -239,6 +250,20 @@ class TestFloat < Test::Unit::TestCase
     assert_equal(-1, (Float::MAX.to_i*2) <=> inf)
     assert_equal(1, (-Float::MAX.to_i*2) <=> -inf)
 
+    bug3609 = '[ruby-core:31470]'
+    def (pinf = Object.new).infinite?; +1 end
+    def (ninf = Object.new).infinite?; -1 end
+    def (fin = Object.new).infinite?; nil end
+    nonum = Object.new
+    assert_equal(0, inf <=> pinf, bug3609)
+    assert_equal(1, inf <=> fin, bug3609)
+    assert_equal(1, inf <=> ninf, bug3609)
+    assert_nil(inf <=> nonum, bug3609)
+    assert_equal(-1, -inf <=> pinf, bug3609)
+    assert_equal(-1, -inf <=> fin, bug3609)
+    assert_equal(0, -inf <=> ninf, bug3609)
+    assert_nil(-inf <=> nonum, bug3609)
+
     assert_raise(ArgumentError) { 1.0 > nil }
     assert_raise(ArgumentError) { 1.0 >= nil }
     assert_raise(ArgumentError) { 1.0 < nil }
@@ -290,11 +315,27 @@ class TestFloat < Test::Unit::TestCase
     assert_raise(FloatDomainError) { inf.ceil }
     assert_raise(FloatDomainError) { inf.round }
     assert_raise(FloatDomainError) { inf.truncate }
+  end
 
+  def test_round_with_precision
     assert_equal(1.100, 1.111.round(1))
     assert_equal(1.110, 1.111.round(2))
     assert_equal(11110.0, 11111.1.round(-1))
     assert_equal(11100.0, 11111.1.round(-2))
+
+    assert_equal(10**300, 1.1e300.round(-300))
+    assert_equal(-10**300, -1.1e300.round(-300))
+    assert_equal(1.0e-300, 1.1e-300.round(300))
+    assert_equal(-1.0e-300, -1.1e-300.round(300))
+
+    bug5227 = '[ruby-core:39093]'
+    assert_equal(42.0, 42.0.round(308), bug5227)
+    assert_equal(1.0e307, 1.0e307.round(2), bug5227)
+
+    assert_raise(TypeError) {1.0.round("4")}
+    assert_raise(TypeError) {1.0.round(nil)}
+    def (prec = Object.new).to_int; 2; end
+    assert_equal(1.0, 0.998.round(prec))
   end
 
   VS = [
@@ -426,16 +467,21 @@ class TestFloat < Test::Unit::TestCase
   def test_Float
     assert_in_delta(0.125, Float("0.1_2_5"), 0.00001)
     assert_in_delta(0.125, "0.1_2_5__".to_f, 0.00001)
-    assert_equal(1, Float(([1] * 10000).join).infinite?)
+    assert_equal(1, suppress_warning {Float(([1] * 10000).join)}.infinite?)
     assert(!Float(([1] * 10000).join("_")).infinite?) # is it really OK?
     assert_raise(ArgumentError) { Float("1.0\x001") }
     assert_equal(15.9375, Float('0xf.fp0'))
-    assert_warn(/malformed value for Float\(\).*?Ruby 1\.9\.3/) { Float('0x') }
-    assert_equal(15.0, Float('0xf'))
-    assert_equal(15.0, Float('0xfp0'))
-    assert_equal(15.0, Float('0xf.p0'))
-    assert_warn(/malformed value for Float\(\).*?Ruby 1\.9\.3/) { Float('0xf.f') }
-    assert_equal(1, Float("1e10_00").infinite?)
+    assert_raise(ArgumentError) { Float('0x') }
+    assert_equal(15, Float('0xf'))
+    assert_equal(15, Float('0xfp0'))
+    assert_raise(ArgumentError) { Float('0xfp') }
+    assert_raise(ArgumentError) { Float('0xf.') }
+    assert_raise(ArgumentError) { Float('0xf.p') }
+    assert_raise(ArgumentError) { Float('0xf.p0') }
+    assert_raise(ArgumentError) { Float('0xf.f') }
+    assert_raise(ArgumentError) { Float('0xf.fp') }
+    assert_equal(Float::INFINITY, Float('0xf.fp1000000000000000'))
+    assert_equal(1, suppress_warning {Float("1e10_00")}.infinite?)
     assert_raise(TypeError) { Float(nil) }
     o = Object.new
     def o.to_f; inf = Float::INFINITY; inf/inf; end
@@ -444,14 +490,7 @@ class TestFloat < Test::Unit::TestCase
 
   def test_invalid_str
     bug4310 = '[ruby-core:34820]'
-    assert_raise(ArgumentError, bug4310) {
-      stress, GC.stress = GC.stress, true
-      begin
-        Float('a'*10000)
-      ensure
-        GC.stress = stress
-      end
-    }
+    assert_raise(ArgumentError, bug4310) {under_gc_stress {Float('a'*10000)}}
   end
 
   def test_num2dbl
