@@ -51,7 +51,7 @@ class TestHash < Test::Unit::TestCase
     assert_equal([], x[22])
     assert_not_same(x[22], x[22])
 
-    x = Hash.new{|h,k| z = k; h[k] = k*2}
+    x = Hash.new{|h,kk| z = kk; h[kk] = kk*2}
     z = 0
     assert_equal(44, x[22])
     assert_equal(22, z)
@@ -113,7 +113,7 @@ class TestHash < Test::Unit::TestCase
     assert_instance_of(@cls, h)
     assert_equal('default', h.default)
     assert_equal('default', h['spurious'])
-    
+
   end
 
   def test_AREF # '[]'
@@ -206,16 +206,20 @@ class TestHash < Test::Unit::TestCase
 
   def test_clone
     for taint in [ false, true ]
-      for frozen in [ false, true ]
-        a = @h.clone
-        a.taint  if taint
-        a.freeze if frozen
-        b = a.clone
+      for untrust in [ false, true ]
+        for frozen in [ false, true ]
+          a = @h.clone
+          a.taint  if taint
+          a.untrust if untrust
+          a.freeze if frozen
+          b = a.clone
 
-        assert_equal(a, b)
-        assert(a.__id__ != b.__id__)
-        assert_equal(a.frozen?, b.frozen?)
-        assert_equal(a.tainted?, b.tainted?)
+          assert_equal(a, b)
+          assert(a.__id__ != b.__id__)
+          assert_equal(a.frozen?, b.frozen?)
+          assert_equal(a.untrusted?, b.untrusted?)
+          assert_equal(a.tainted?, b.tainted?)
+        end
       end
     end
   end
@@ -286,18 +290,28 @@ class TestHash < Test::Unit::TestCase
     assert_equal(base.size, n)
   end
 
+  def test_keep_if
+    h = {1=>2,3=>4,5=>6}
+    assert_equal({3=>4,5=>6}, h.keep_if {|k, v| k + v >= 7 })
+    h = {1=>2,3=>4,5=>6}
+    assert_equal({1=>2,3=>4,5=>6}, h.keep_if{true})
+  end
+
   def test_dup
     for taint in [ false, true ]
-      for frozen in [ false, true ]
-        a = @h.dup
-        a.taint  if taint
-        a.freeze if frozen
-        b = a.dup
+      for untrust in [ false, true ]
+        for frozen in [ false, true ]
+          a = @h.dup
+          a.taint  if taint
+          a.freeze if frozen
+          b = a.dup
 
-        assert_equal(a, b)
-        assert(a.__id__ != b.__id__)
-        assert_equal(false, b.frozen?)
-        assert_equal(a.tainted?, b.tainted?)
+          assert_equal(a, b)
+          assert(a.__id__ != b.__id__)
+          assert_equal(false, b.frozen?)
+          assert_equal(a.tainted?, b.tainted?)
+          assert_equal(a.untrusted?, b.untrusted?)
+        end
       end
     end
   end
@@ -359,8 +373,6 @@ class TestHash < Test::Unit::TestCase
   end
 
   def test_fetch
-    assert_raise(KeyError) { @cls[].fetch(1) }
-    assert_raise(KeyError) { @h.fetch('gumby') }
     assert_equal('gumbygumby', @h.fetch('gumby') {|k| k * 2 })
     assert_equal('pokey', @h.fetch('gumby', 'pokey'))
 
@@ -369,7 +381,15 @@ class TestHash < Test::Unit::TestCase
     assert_equal('nil', @h.fetch(nil))
   end
 
-  def test_key?
+  def test_fetch_error
+    assert_raise(KeyError) { @cls[].fetch(1) }
+    assert_raise(KeyError) { @h.fetch('gumby') }
+    e = assert_raise(KeyError) { @h.fetch('gumby'*20) }
+    assert_match(/key not found: "gumbygumby/, e.message)
+    assert_match(/\.\.\.\z/, e.message)
+  end
+
+  def test_key2?
     assert(!@cls[].key?(1))
     assert(!@cls[].key?(nil))
     assert(@h.key?(nil))
@@ -532,7 +552,7 @@ class TestHash < Test::Unit::TestCase
 
   def test_shift
     h = @h.dup
-    
+
     @h.length.times {
       k, v = h.shift
       assert(@h.key?(k))
@@ -599,6 +619,13 @@ class TestHash < Test::Unit::TestCase
     assert_equal([3,4], a.delete([3,4]))
     assert_equal([5,6], a.delete([5,6]))
     assert_equal(0, a.length)
+
+    h = @cls[ 1=>2, 3=>4, 5=>6 ]
+    h.taint
+    h.untrust
+    a = h.to_a
+    assert_equal(true, a.tainted?)
+    assert_equal(true, a.untrusted?)
   end
 
   def test_to_hash
@@ -613,6 +640,7 @@ class TestHash < Test::Unit::TestCase
     assert_equal(h.inspect, h.to_s)
     h = @cls[]
     assert_equal(h.inspect, h.to_s)
+  ensure
     $, = nil
   end
 
@@ -633,7 +661,7 @@ class TestHash < Test::Unit::TestCase
     assert_equal(hb, h2)
   end
 
-  def test_value?
+  def test_value2?
     assert(!@cls[].value?(1))
     assert(!@cls[].value?(nil))
     assert(@h.value?(nil))
@@ -688,14 +716,14 @@ class TestHash < Test::Unit::TestCase
   end
 
   def test_default_proc
-    h = Hash.new {|h, k| h + k + "baz" }
+    h = Hash.new {|hh, k| hh + k + "baz" }
     assert_equal("foobarbaz", h.default_proc.call("foo", "bar"))
     h = {}
     assert_nil(h.default_proc)
   end
 
   def test_shift2
-    h = Hash.new {|h, k| :foo }
+    h = Hash.new {|hh, k| :foo }
     h[1] = 2
     assert_equal([1, 2], h.shift)
     assert_equal(:foo, h.shift)
@@ -721,6 +749,14 @@ class TestHash < Test::Unit::TestCase
     assert_equal({3=>4,5=>6}, {1=>2,3=>4,5=>6}.select {|k, v| k + v >= 7 })
   end
 
+  def test_select!
+    h = {1=>2,3=>4,5=>6}
+    assert_equal(h, h.select! {|k, v| k + v >= 7 })
+    assert_equal({3=>4,5=>6}, h)
+    h = {1=>2,3=>4,5=>6}
+    assert_equal(nil, h.select!{true})
+  end
+
   def test_clear2
     assert_equal({}, {1=>2,3=>4,5=>6}.clear)
     h = {1=>2,3=>4,5=>6}
@@ -733,6 +769,13 @@ class TestHash < Test::Unit::TestCase
     h2 = {}
     h2.replace h1
     assert_equal(:foo, h2[0])
+
+    assert_raise(ArgumentError) { h2.replace() }
+    assert_raise(TypeError) { h2.replace(1) }
+    h2.freeze
+    assert_raise(ArgumentError) { h2.replace() }
+    assert_raise(RuntimeError) { h2.replace(h1) }
+    assert_raise(RuntimeError) { h2.replace(42) }
   end
 
   def test_size2
@@ -752,10 +795,6 @@ class TestHash < Test::Unit::TestCase
     assert(h1 != h2)
     h1 = {1=>2}; h2 = {1=>4}
     assert(h1 != h2)
-
-    h1 = {}; h1[h1] = h1; h1.rehash
-    h2 = {}; h2[h2] = h2; h2.rehash
-    assert(h1 != h2)
   end
 
   def test_eql
@@ -770,6 +809,9 @@ class TestHash < Test::Unit::TestCase
 
   def test_hash2
     assert_kind_of(Integer, {}.hash)
+    h = {1=>2}
+    h.shift
+    assert_equal({}.hash, h.hash, '[ruby-core:38650]')
   end
 
   def test_update2
@@ -831,7 +873,51 @@ class TestHash < Test::Unit::TestCase
     assert_nil(h["foo"])
   end
 
+  class ObjWithHash
+    def initialize(value, hash)
+      @value = value
+      @hash = hash
+    end
+    attr_reader :value, :hash
+
+    def eql?(other)
+      @value == other.value
+    end
+  end
+
   def test_hash_hash
     assert_equal({0=>2,11=>1}.hash, {11=>1,0=>2}.hash)
+    o1 = ObjWithHash.new(0,1)
+    o2 = ObjWithHash.new(11,1)
+    assert_equal({o1=>1,o2=>2}.hash, {o2=>2,o1=>1}.hash)
+  end
+
+  def test_hash_bignum_hash
+    x = 2<<(32-3)-1
+    assert_equal({x=>1}.hash, {x=>1}.hash)
+    x = 2<<(64-3)-1
+    assert_equal({x=>1}.hash, {x=>1}.hash)
+
+    o = Object.new
+    def o.hash; 2 << 100; end
+    assert_equal({x=>1}.hash, {x=>1}.hash)
+  end
+
+  def test_hash_poped
+    assert_nothing_raised { eval("a = 1; {a => a}; a") }
+  end
+
+  def test_recursive_key
+    h = {}
+    assert_nothing_raised { h[h] = :foo }
+    h.rehash
+    assert_equal(:foo, h[h])
+  end
+
+  def test_inverse_hash
+    feature4262 = '[ruby-core:34334]'
+    [{1=>2}, {123=>"abc"}].each do |h|
+      assert_not_equal(h.hash, h.invert.hash, feature4262)
+    end
   end
 end
