@@ -232,6 +232,30 @@ class TestM17N < Test::Unit::TestCase
     Encoding.default_external = orig_ext
   end
 
+  def test_utf_16_32_inspect
+    str = "\u3042"
+    %w/UTF-16 UTF-32/.each do |enc|
+      %w/BE LE/.each do |endian|
+        s = str.encode(enc + endian)
+        # When a UTF-16/32 string doesn't have a BOM,
+        # inspect as a dummy encoding string.
+        assert_equal(s.dup.force_encoding("ISO-2022-JP").inspect,
+                     s.dup.force_encoding(enc).inspect)
+      end
+    end
+
+    str = "\uFEFF\u3042"
+    %w/UTF-16 UTF-32/.each do |enc|
+      %w/BE LE/.each do |endian|
+        s = str.encode(enc + endian)
+        # When a UTF-16/32 string doesn't have a BOM,
+        # inspect as a dummy encoding string.
+        assert_equal(s.inspect,
+                     s.dup.force_encoding(enc).inspect)
+      end
+    end
+  end
+
   def test_str_dump
     [
       e("\xfe"),
@@ -538,6 +562,14 @@ class TestM17N < Test::Unit::TestCase
     assert(r.fixed_encoding?)
     assert_match(r, "\xa4\xa2".force_encoding("euc-jp"))
 
+    r = /\p{AsciI}/e
+    assert(r.fixed_encoding?)
+    assert_match(r, "a".force_encoding("euc-jp"))
+
+    r = /\p{hiraganA}/e
+    assert(r.fixed_encoding?)
+    assert_match(r, "\xa4\xa2".force_encoding("euc-jp"))
+
     r = eval('/\u{3042}\p{Hiragana}/'.force_encoding("euc-jp"))
     assert(r.fixed_encoding?)
     assert_equal(Encoding::UTF_8, r.encoding)
@@ -779,15 +811,15 @@ class TestM17N < Test::Unit::TestCase
   end
 
   def test_sprintf_p
-    enc = "".inspect.encoding
-    asc = Encoding::US_ASCII
     Encoding.list.each do |e|
       format = "%p".force_encoding(e)
       ['', 'a', "\xC2\xA1", "\x00"].each do |s|
         s.force_encoding(e)
-        assert_strenc(s.inspect, e.ascii_compatible? && enc == asc ? e : enc, format % s)
+        enc = (''.force_encoding(e) + s.inspect).encoding
+        assert_strenc(s.inspect, enc, format % s)
       end
       s = "\xC2\xA1".force_encoding(e)
+      enc = ('' + s.inspect).encoding
       assert_strenc('%10s' % s.inspect, enc, "%10p" % s)
     end
   end
@@ -901,6 +933,10 @@ class TestM17N < Test::Unit::TestCase
   def test_aset
     s = e("\xa3\xb0\xa3\xb1\xa3\xb2\xa3\xb3\xa3\xb4")
     assert_raise(Encoding::CompatibilityError){s["\xb0\xa3"] = "foo"}
+
+    a = ua("a")
+    a[/a/] = u("")
+    assert_equal Encoding::US_ASCII, a.encoding
   end
 
   def test_str_center
@@ -926,6 +962,7 @@ class TestM17N < Test::Unit::TestCase
 
     assert_equal("X\u3042\u3044X", "A\u3042\u3044\u3046".tr("^\u3042\u3044", "X"))
     assert_equal("\u3042\u3046" * 100, ("\u3042\u3044" * 100).tr("\u3044", "\u3046"))
+    assert_equal("Y", "\u3042".tr("^X", "Y"))
   end
 
   def test_tr_s
@@ -1054,7 +1091,6 @@ class TestM17N < Test::Unit::TestCase
     assert_equal(false, s.ascii_only?, "[ruby-core:14566] reported by Sam Ruby")
 
     s = "abc".force_encoding(Encoding::ASCII_8BIT)
-    t = s.gsub(/b/, "\xa1\xa1".force_encoding("euc-jp"))
     assert_equal(Encoding::ASCII_8BIT, s.encoding)
 
     assert_raise(Encoding::CompatibilityError) {
@@ -1066,8 +1102,8 @@ class TestM17N < Test::Unit::TestCase
     s = e("\xa3\xb0\xa3\xb1\xa3\xb2\xa3\xb3\xa3\xb4")
     assert_equal(e("\xa3\xb0z\xa3\xb2\xa3\xb3\xa3\xb4"), s.gsub(/\xa3\xb1/e, "z"))
 
-    assert_equal(Encoding::EUC_JP, (a("").gsub(//) { e("") }.encoding))
-    assert_equal(Encoding::EUC_JP, (a("a").gsub(/a/) { e("") }.encoding))
+    assert_equal(Encoding::ASCII_8BIT, (a("").gsub(//) { e("") }.encoding))
+    assert_equal(Encoding::ASCII_8BIT, (a("a").gsub(/a/) { e("") }.encoding))
   end
 
   def test_end_with
@@ -1101,6 +1137,7 @@ class TestM17N < Test::Unit::TestCase
 
   def test_str_concat
     assert_equal(1, "".concat(0xA2).size)
+    assert_equal(Encoding::ASCII_8BIT, "".force_encoding("US-ASCII").concat(0xA2).encoding)
     assert_equal("A\x84\x31\xA4\x39".force_encoding("GB18030"),
                  "A".force_encoding("GB18030") << 0x8431A439)
   end
@@ -1143,7 +1180,7 @@ class TestM17N < Test::Unit::TestCase
 
   def test_symbol_op
     ops = %w"
-      .. ... + - +(binary) -(binary) * / % ** +@ -@ | ^ & ! <=> > >= < <= ==
+      .. ... + - * / % ** +@ -@ | ^ & ! <=> > >= < <= ==
       === != =~ !~ ~ ! [] []= << >> :: `
     "
     ops.each do |op|
@@ -1156,6 +1193,18 @@ class TestM17N < Test::Unit::TestCase
       assert_equal([b].pack("C"), b.chr)
     }
     assert_equal("\x84\x31\xA4\x39".force_encoding("GB18030"), 0x8431A439.chr("GB18030"))
+    e = assert_raise(RangeError) {
+      2206368128.chr(Encoding::UTF_8)
+    }
+    assert_not_match(/-\d+ out of char range/, e.message)
+
+    assert_raise(RangeError){ 0x80.chr("US-ASCII") }
+    assert_raise(RangeError){ 0x80.chr("SHIFT_JIS") }
+    assert_raise(RangeError){ 0xE0.chr("SHIFT_JIS") }
+    assert_raise(RangeError){ 0x100.chr("SHIFT_JIS") }
+    assert_raise(RangeError){ 0xA0.chr("EUC-JP") }
+    assert_raise(RangeError){ 0x100.chr("EUC-JP") }
+    assert_raise(RangeError){ 0xA1A0.chr("EUC-JP") }
   end
 
   def test_marshal
@@ -1320,6 +1369,14 @@ class TestM17N < Test::Unit::TestCase
     assert_equal(true, s.valid_encoding?)
     s << "\xff".force_encoding("utf-16be")
     assert_equal(false, s.valid_encoding?, bug4018)
+
+    bug6190 = '[ruby-core:43557]'
+    s = "\xe9"
+    s = s.encode("utf-8", "utf-8")
+    assert_equal(false, s.valid_encoding?, bug6190)
+    s = "\xe9"
+    s.encode!("utf-8", "utf-8")
+    assert_equal(false, s.valid_encoding?, bug6190)
   end
 
   def test_getbyte
@@ -1368,7 +1425,7 @@ class TestM17N < Test::Unit::TestCase
     Encoding.list.each do |enc|
       next if enc.dummy?
       strs = strings.map {|s| s.encode(enc)} rescue next
-      yield *strs
+      yield(*strs)
     end
   end
 end
