@@ -1,20 +1,20 @@
-#!./miniruby -s
+#!./miniruby -sI.
 
 $name = $library = $description = nil
 
 module RbConfig
-  autoload :CONFIG, "rbconfig"
+  autoload :CONFIG, "./rbconfig"
 end
 
 class Exports
-  @subclass = []
+  @@subclass = []
   def self.inherited(klass)
-    @subclass << [/#{klass.name.sub(/.*::/, '').downcase}/i, klass]
+    @@subclass << [/#{klass.name.sub(/.*::/, '').downcase}/i, klass]
   end
 
   def self.create(*args, &block)
     platform = RUBY_PLATFORM
-    pat, klass = @subclass.find {|p, k| p =~ platform}
+    pat, klass = @@subclass.find {|p, k| p =~ platform}
     unless klass
       raise ArgumentError, "unsupported platform: #{platform}"
     end
@@ -36,6 +36,7 @@ class Exports
   def initialize(objs)
     syms = {}
     winapis = {}
+    syms["ruby_sysinit_real"] = "ruby_sysinit"
     each_export(objs) do |internal, export|
       syms[internal] = export
       winapis[$1] = internal if /^_?(rb_w32_\w+)(?:@\d+)?$/ =~ internal
@@ -51,8 +52,8 @@ class Exports
       end
     end
     syms["NtInitialize"] ||= "ruby_sysinit" if syms["ruby_sysinit"]
-    syms["rb_w32_vsnprintf"] ||= "vsnprintf"
-    syms["rb_w32_snprintf"] ||= "snprintf"
+    syms["rb_w32_vsnprintf"] ||= "ruby_vsnprintf"
+    syms["rb_w32_snprintf"] ||= "ruby_snprintf"
     @syms = syms
   end
 
@@ -110,7 +111,8 @@ class Exports::Mswin < Exports
           next unless l.sub!(/.*?\s(\(\)\s+)?External\s+\|\s+/, '')
           is_data = !$1
           if noprefix or /^[@_]/ =~ l
-            next if /(?!^)@.*@/ =~ l || /@[[:xdigit:]]{16}$/ =~ l
+            next if /(?!^)@.*@/ =~ l || /@[[:xdigit:]]{8,16}$/ =~ l ||
+              /^_?(?:Init_|.*_threadptr_|DllMain\b)/ =~ l
             l.sub!(/^[@_]/, '') if /@\d+$/ !~ l
           elsif !l.sub!(/^(\S+) \([^@?\`\']*\)$/, '\1')
             next
@@ -128,7 +130,7 @@ class Exports::Mswin < Exports
   end
 end
 
-class Exports::Mingw < Exports
+class Exports::Cygwin < Exports
   def self.nm
     @@nm ||= RbConfig::CONFIG["NM"]
   end
@@ -142,9 +144,19 @@ class Exports::Mingw < Exports
   end
 
   def each_export(objs)
+    symprefix = RbConfig::CONFIG["SYMBOL_PREFIX"]
+    symprefix.strip! if symprefix
+    re = /\s(?:(T)|[[:upper:]])\s#{symprefix}((?!Init_|.*_threadptr_|DllMain\b).*)$/
     objdump(objs) do |l|
-      yield $2, !$1 if /\s(?:(T)|[[:upper:]])\s_((?!Init_).*)$/ =~ l
+      next if /@.*@/ =~ l
+      yield $2, !$1 if re =~ l
     end
+  end
+end
+
+class Exports::Mingw < Exports::Cygwin
+  def each_export(objs)
+    super
     yield "strcasecmp", "_stricmp"
     yield "strncasecmp", "_strnicmp"
   end
